/*

 65c816 opcodes 0x00-0x0F:
    00: BRK s       2,8
    01: ORA (d,x)   2,6
    02: COP s       2,8
    03: ORA d,s     2,4
    04: TSB d       2,5
    05: ORA d       2,3
    06: ASL d       2,5
    07: ORA [d]     2,6
    08: PHP s       1,3
    09: ORA #       2,2
    0A: ASL a       1,2
    0B: PHD s       1,4
    0C: TSB a       3,6
    0D: ORA a       3,4
    0E: ASL a       3,6
    0F: ORA al      4,5
*/

.macro NativeInterrupt vector,offset
    movb R_PB,%al
    E0_PUSH_B
    GET_PC %eax
    addl $2,%eax
    E0_PUSH_W
    E0_SETUPFLAGS               // Put flags into true 65c816 format
    E0_PUSH_B

    movl \vector,_NativePC      // PC = Native mode vector
    movl \offset,%eax
    movb $0,_PB                 // Setup bank
    movl %eax,R_PBOffset    // Offset of bank 0 ROM
    SET_FLAG SNES_FLAG_I    // Disable IRQs
    addl %eax,_NativePC
    movl _Read_BankMapping,%eax
    movb $4,CycleTable  // SlowROM bank
    CLR_FLAG SNES_FLAG_D    // Disable decimal mode
    movl %eax,R_PB_Read // Bank 0 read through memory mapper
    jmp CPU_RETURN
.endm

.macro Emu6502Interrupt vector,offset
    GET_PC %eax
    addl $2,%eax
    E1_PUSH_W
    E1_SETUPFLAGS               // Put flags into true 65c816 format
    E1_PUSH_B

    movl \vector,_NativePC      // PC = Native mode vector
    movl \offset,%eax
    movb $0,_PB                 // Setup bank
    movl %eax,R_PBOffset    // Offset of bank 0 ROM
    SET_FLAG SNES_FLAG_I    // Disable IRQs
    addl %eax,_NativePC
    movl _Read_BankMapping,%eax
    movb $4,CycleTable  // SlowROM bank
    CLR_FLAG SNES_FLAG_D    // Disable decimal mode
    movl %eax,R_PB_Read // Bank 0 read through memory mapper
    jmp CPU_RETURN
.endm

/* 00 */

ALIGNC
E0_BRK:
    NativeInterrupt R_BRK_Nvector,R_BRK_Noffset

ALIGNC
E1_BRK:
    Emu6502Interrupt R_IRQ_Evector,R_IRQ_Eoffset

/* 01 */
EM_ORA Od_xO,Direct_Indexed_Indirect
E0_ORA Od_xO,Direct_Indexed_Indirect

/* 02 */

ALIGNC
E0_COP:
    NativeInterrupt R_COP_Nvector,R_COP_Noffset

ALIGNC
E1_COP:
    Emu6502Interrupt R_COP_Evector,R_COP_Eoffset

/* 03 */

EM_ORA d_s,Stack_Relative
E0_ORA d_s,Stack_Relative

/* 04 */
EM_TSB d,Direct
E0_TSB d,Direct

/* 05 */
EM_ORA d,Direct
E0_ORA d,Direct

/* 06 */
EM_ASL d,Direct
E0_ASL d,Direct

/* 07 */
EM_ORA IdI,Direct_Indirect_Long
E0_ORA IdI,Direct_Indirect_Long

/* 08 */

ALIGNC
E0_PHP:
    incl _NativePC
    E0_SETUPFLAGS
    E0_PUSH_B
    jmp CPU_RETURN
ALIGNC
E1_PHP:
    incl _NativePC
    E1_SETUPFLAGS
    E1_PUSH_B
    jmp CPU_RETURN

/* 09 */
EM_ORA i,Immediate
E0_ORA i,Immediate

/* 0A */

ALIGNC
EM_SLA:
    incl _NativePC
    shlb $1,R_A
    STORE_FLAGS (SNES_FLAG_NZC)
    jmp CPU_RETURN
ALIGNC
E0_SLA:
    incl _NativePC
    shlw $1,R_A
    STORE_FLAGS (SNES_FLAG_NZC)
    jmp CPU_RETURN

/* 0B */
ALIGNC
E0_PHD:
    incl _NativePC
    movl R_D,%eax
    E0_PUSH_W
    jmp CPU_RETURN
ALIGNC
E1_PHD:
    incl _NativePC
    movl R_D,%eax
    E1_PUSH_W
    jmp CPU_RETURN

/* 0C */
EM_TSB a,Absolute
E0_TSB a,Absolute

/* 0D */
EM_ORA a,Absolute
E0_ORA a,Absolute

/* 0E */
EM_ASL a,Absolute
E0_ASL a,Absolute

/* 0F */
EM_ORA al,Absolute_Long
E0_ORA al,Absolute_Long
