/*

 65c816 opcodes 0x40-0x4F:
    40: RTI s       1,7
    41: EOR (d,x)   2,6
    42: WDM         2,2
    43: EOR d,s     2,4
    44: MVP xya     3,7
    45: EOR d       2,3
    46: LSR d       2,5
    47: EOR [d]     2,6
    48: PHA s       1,3
    49: EOR #       2,2
    4A: LSR a       1,2
    4B: PHK s       1,3
    4C: JMP a       3,3
    4D: EOR a       3,4
    4E: LSR a       3,6
    4F: EOR al      4,5
*/

/* 40 */

ALIGNC
E0_RTI:
    E0_PULL_B
    E0_RESTOREFLAGS
    E0_PULL_W
    movl %eax,_NativePC
    E0_PULL_B
    xorl %ebx,%ebx
    movb %al,R_PB
    movb %al,%bl
    testb %al,MEMSEL    // Check bus speed against PB
    movb $5,%al
    jnz 0f
    movb $4,%al
0:  addb %ah,%ah
    movb %al,CycleTable // Update bus speed
    movl _Read_BankMapping(,%ebx,4),%eax
    adcl %ebx,%ebx
    movl %eax,R_PB_Read
    movl _Read_BankOffset(,%ebx,4),%eax
    addl %eax,_NativePC
    movl %eax,R_PBOffset

    JUMP_FLAG SNES_FLAG_I,0f
    movb IRQ_pin,%al
    testb %al,%al
    jz  0f
    SAVE_PC %ebx
    addl $60,_Cycles    // IRQ processing: 2 IO + 6 bank0
    call E0_IRQ
    LOAD_PC
0:
    SET_TABLE_MX

ALIGNC
E1_RTI:
    E1_PULL_B
    E1_RESTOREFLAGS
    E1_PULL_W
    movl %eax,_NativePC
    xorl %ebx,%ebx
    addb %ah,%ah
    movb R_PB,%bl
    adcl %ebx,%ebx
    movl _Read_BankOffset(,%ebx,4),%eax
    addl %eax,_NativePC
    movl %eax,R_PBOffset

    JUMP_FLAG SNES_FLAG_I,CPU_RETURN
    movb IRQ_pin,%al
    testb %al,%al
    jz  CPU_RETURN
    SAVE_PC %ebx
    addl $52,_Cycles    // IRQ processing: 2 IO + 5 bank0
    call E1_IRQ
    LOAD_PC
    jmp CPU_RETURN

/* 41 */
EM_EOR Od_xO,Direct_Indexed_Indirect
E0_EOR Od_xO,Direct_Indexed_Indirect

/* 42 */
ALIGNC
ALL_WDM:
    addl $2,_NativePC
    jmp CPU_RETURN

/* 43 */
EM_EOR d_s,Stack_Relative
E0_EOR d_s,Stack_Relative

/* 44 */

ALIGNC
E0_MVP:
    pushl %ebx          // Cycles per instruction iteration
0:  movb 1(_NativePC),%al   // Dest bank
    movb 2(_NativePC),%bl   // Src bank
    movb %al,R_DB
    movw R_X,%ax        // Src address
    shll $16,%ebx
    addl %eax,%ebx
    decl %eax
    movw %ax,R_X
    GET_BYTE
    movl R_DB_Shifted,%ebx
    movw R_Y,%bx        // Dest address
    SET_BYTE
    movl R_A,%eax
    decl %ebx
    subl $1,%eax
    movw %bx,R_Y
    jc  0f
    cmpl _Cycles,R_EventTrip
    jae 1f
    movw %ax,R_A
    addl (%esp),_Cycles
    xorl %eax,%eax
    xorl %ebx,%ebx
    jmp 0b
ALIGNC
0:  addl $3,_NativePC
1:  movw %ax,R_A
    popl %edi
    jmp CPU_RETURN
	
ALIGNC
EX_MVP:
    pushl %ebx          // Cycles per instruction iteration
0:  movb 1(_NativePC),%al   // Dest bank
    movb 2(_NativePC),%bl   // Src bank
    movb %al,R_DB
    movb R_X,%al        // Src address
    shll $16,%ebx
    addl %eax,%ebx
    decl %eax
    movb %al,R_X
    GET_BYTE
    movl R_DB_Shifted,%ebx
    movb R_Y,%bl        // Dest address
    SET_BYTE
    movl R_A,%eax
    decl %ebx
    subl $1,%eax
    movb %bl,R_Y
    jc  0f
    cmpl _Cycles,R_EventTrip
    jae 1f
    movw %ax,R_A
    addl (%esp),_Cycles
    xorl %eax,%eax
    xorl %ebx,%ebx
    jmp 0b
ALIGNC
0:  addl $3,_NativePC
1:  movw %ax,R_A
    popl %edi
    jmp CPU_RETURN

/* 45 */
EM_EOR d,Direct
E0_EOR d,Direct

/* 46 */
EM_LSR d,Direct
E0_LSR d,Direct

/* 47 */
EM_EOR IdI,Direct_Indirect_Long
E0_EOR IdI,Direct_Indirect_Long

/* 48 */

ALIGNC
E0_PHA:
    incl _NativePC
    movl R_A,%eax
    E0_PUSH_W
    jmp CPU_RETURN

ALIGNC
EM_PHA:
    incl _NativePC
    movl R_A,%eax
    E0_PUSH_B
    jmp CPU_RETURN

ALIGNC
E1_PHA:
    incl _NativePC
    movl R_A,%eax
    E1_PUSH_B
    jmp CPU_RETURN

/* 49 */
EM_EOR i,Immediate
E0_EOR i,Immediate

/* 4A */

ALIGNC
EM_SRA:
    incl _NativePC
    shrb $1,R_A
    STORE_FLAGS (SNES_FLAG_NZC)
    jmp CPU_RETURN
ALIGNC
E0_SRA:
    incl _NativePC
    shrw $1,R_A
    STORE_FLAGS (SNES_FLAG_NZC)
    jmp CPU_RETURN

/* 4B */

ALIGNC
E0_PHK:				/* Push PB(K) onto stack */
    incl _NativePC
    movb R_PB,%al
    E0_PUSH_B
    jmp CPU_RETURN

ALIGNC
E1_PHK:
    incl _NativePC
    movb R_PB,%al
    E1_PUSH_B
    jmp CPU_RETURN

/* 4C */

ALIGNC
ALL_JMP_a:
    ADDR_Absolute_JMP
    movl %eax,_NativePC
    movb R_PB,%bl
    addb %ah,%ah
    adcl %ebx,%ebx
    movl _Read_BankOffset(,%ebx,4),%eax
    addl %eax,_NativePC
    movl %eax,R_PBOffset
    jmp CPU_RETURN

/* 4D */
EM_EOR a,Absolute
E0_EOR a,Absolute

/* 4E */
EM_LSR a,Absolute
E0_LSR a,Absolute

/* 4F */
EM_EOR al,Absolute_Long
E0_EOR al,Absolute_Long
