/*

 65c816 opcodes 0x60-0x6F:
    60: RTS s       1,6
    61: ADC (d,x)   2,6
    62: PER s       3,6
    63: ADC d,s     2,4
    64: STZ d       2,3
    65: ADC d       2,3
    66: ROR d       2,5
    67: ADC [d]     2,6
    68: PLA s       1,4
    69: ADC #       2,2
    6A: ROR a       1,2
    6B: RTL s       1,6
    6C: JMP (a)     3,5
    6D: ADC a       3,4
    6E: ROR a       3,6
    6F: ADC al      4,5
*/

/* 60 */

ALIGNC
E0_RTS:
    E0_PULL_W
    incw %ax                // This is a MAD processor!
    xorl %ebx,%ebx
    movl %eax,_NativePC
    movb R_PB,%bl
    addb %ah,%ah
    adcl %ebx,%ebx
    movl _Read_BankOffset(,%ebx,4),%eax
    addl %eax,_NativePC
    movl %eax,R_PBOffset
    jmp CPU_RETURN
ALIGNC
E1_RTS:
    E1_PULL_W
    incw %ax                // This is a MAD processor!
    xorl %ebx,%ebx
    movl %eax,_NativePC
    movb R_PB,%bl
    addb %ah,%ah
    adcl %ebx,%ebx
    movl _Read_BankOffset(,%ebx,4),%eax
    addl %eax,_NativePC
    movl %eax,R_PBOffset
    jmp CPU_RETURN

/* 61 */
EM_ADC Od_xO,Direct_Indexed_Indirect
E0_ADC Od_xO,Direct_Indexed_Indirect

/* 62 */

ALIGNC
E0_PER:
    movb 1(_NativePC),%al
    movb 2(_NativePC),%ah
    addl $3,_NativePC
    GET_PC %ebx
    addl %ebx,%eax
	E0_PUSH_W
    jmp CPU_RETURN

ALIGNC
E1_PER:
    movb 1(_NativePC),%al
    movb 2(_NativePC),%ah
    addl $3,_NativePC
    GET_PC %ebx
    addl %ebx,%eax
	E1_PUSH_W
    jmp CPU_RETURN

/* 63 */
EM_ADC d_s,Stack_Relative
E0_ADC d_s,Stack_Relative

/* 64 */
EM_STZ d,Direct
E0_STZ d,Direct

/* 65 */
EM_ADC d,Direct
E0_ADC d,Direct

/* 66 */
EM_ROR d,Direct
E0_ROR d,Direct

/* 67 */
EM_ADC IdI,Direct_Indirect_Long
E0_ADC IdI,Direct_Indirect_Long

/* 68 */

ALIGNC
E0_PLA:
    incl _NativePC
    E0_PULL_W
    movl %eax,R_A
    testw %ax,%ax
    STORE_FLAGS (SNES_FLAG_NZ)
    jmp CPU_RETURN
ALIGNC
EM_PLA:
    incl _NativePC
    E0_PULL_B
    movb %al,R_A
    testb %al,%al
    STORE_FLAGS (SNES_FLAG_NZ)
    jmp CPU_RETURN
ALIGNC
E1_PLA:
    incl _NativePC
    E1_PULL_B
    movb %al,R_A
    testb %al,%al
    STORE_FLAGS (SNES_FLAG_NZ)
    jmp CPU_RETURN

/* 69 */
EM_ADC i,Immediate
E0_ADC i,Immediate

/* 6A */    // Fixed 0.25: RRA8 didn't set NZ flags correctly

ALIGNC
EM_RRA:
    shrb $1,_P_B    //MAKE_CARRY
    movb R_A,%al
    incl _NativePC
    rcrb $1,%al
    adcb _P_B,_P_B  //STORE_FLAGS (SNES_FLAG_C)
    testb %al,%al
    STORE_FLAGS (SNES_FLAG_NZ)
    movb %al,R_A
    jmp CPU_RETURN

ALIGNC
E0_RRA:
    shrb $1,_P_B    //MAKE_CARRY
    movl R_A,%edi
    incl _NativePC
    rcrw $1,%di
    adcb _P_B,_P_B  //STORE_FLAGS (SNES_FLAG_C)
    testw %di,%di
    STORE_FLAGS (SNES_FLAG_NZ)
    movl %edi,R_A
    jmp CPU_RETURN

/* 6B */

ALIGNC
E0_RTL:
    E0_PULL_W
    incw %ax                // This is a MAD processor!
    movl %eax,_NativePC
    E0_PULL_B
    xorl %ebx,%ebx
    movb %al,R_PB
    movb %al,%bl
    testb %al,MEMSEL    // Check bus speed against PB
    movb $5,%al
    jnz 0f
    movb $4,%al
0:  addb %ah,%ah
    movb %al,CycleTable // Update bus speed
    movl _Read_BankMapping(,%ebx,4),%eax
    adcl %ebx,%ebx
    movl %eax,R_PB_Read
    movl _Read_BankOffset(,%ebx,4),%eax
    addl %eax,_NativePC
    movl %eax,R_PBOffset
    jmp CPU_RETURN
ALIGNC
E1_RTL:
    E1_PULL_W
    incw %ax                // This is a MAD processor!
    movl %eax,_NativePC
    E1_PULL_B
    xorl %ebx,%ebx
    movb %al,R_PB
    movb %al,%bl
    testb %al,MEMSEL    // Check bus speed against PB
    movb $5,%al
    jnz 0f
    movb $4,%al
0:  addb %ah,%ah
    movb %al,CycleTable // Update bus speed
    movl _Read_BankMapping(,%ebx,4),%eax
    adcl %ebx,%ebx
    movl %eax,R_PB_Read
    movl _Read_BankOffset(,%ebx,4),%eax
    addl %eax,_NativePC
    movl %eax,R_PBOffset
    jmp CPU_RETURN

/* 6C */

/*
 JMP Absolute Indirect
 Returns PCH in %al
  PC = (a16)
*/
ALIGNC
ALL_JMP_OaO:
    movb 1(_NativePC),%bl
    movb 2(_NativePC),%bh
    GET_BYTE
    incl %ebx
    movb %al,%ah
    GET_BYTE
    xchgb %al,%ah
    xorl %ebx,%ebx
    movl %eax,_NativePC
    movb _PB,%bl
    addb %ah,%ah
    adcl %ebx,%ebx
    movl _Read_BankOffset(,%ebx,4),%eax
    addl %eax,_NativePC
    movl %eax,_PBOffset
    jmp CPU_RETURN

/* 6D */
EM_ADC a,Absolute
E0_ADC a,Absolute

/* 6E */
EM_ROR a,Absolute
E0_ROR a,Absolute

/* 6F */
EM_ADC al,Absolute_Long
E0_ADC al,Absolute_Long
