/*
 SNEeSe 65c816 CPU emulation core
 Originally written by Savoury SnaX (Not quite sure if I like AT&T)
 Maintained/rewritten by Charles Bilyue'

 Compile under DJGPP (GNU-AS)
  NOTE - This must have .S (capital) or the # stuff wont be pre-processed!!!

 This file contains:
  CPU opcode macros

 Examples of usage:
  EM_LDA Direct = 8-bit LDA using "direct" address mode
  BFC SNES_FLAG_C = Branch if carry flag clear (BCC)

*/

.macro EM_ADC label,addr
ALIGNC
EM_ADC_\label:
    READ8_\addr
    CLR_FLAG(SNES_FLAG_V)
    JUMP_FLAG SNES_FLAG_D,0f
    MAKE_CARRY
    adcb %al,R_A
    lahf
    jno 2f
    SET_FLAG(SNES_FLAG_V)
2:  STORE_FLAGS_FROM SNES_FLAG_NZC,%ah
    jmp CPU_RETURN

ALIGNC
0:  MAKE_CARRY
    adcb R_A,%al
    daa
    movb %al,R_A
    lahf
    jno 2f
1:  SET_FLAG(SNES_FLAG_V)
2:  STORE_FLAGS_FROM SNES_FLAG_NZC,%ah
    jmp CPU_RETURN
.endm

.macro E0_ADC label,addr
ALIGNC
E0_ADC_\label:
    READ16_\addr
    CLR_FLAG(SNES_FLAG_V)
    JUMP_FLAG SNES_FLAG_D,0f
    MAKE_CARRY
    adcw %ax,R_A
    lahf
    jno 1f
    SET_FLAG(SNES_FLAG_V)
1:  STORE_FLAGS_FROM SNES_FLAG_NZC,%ah
    jmp CPU_RETURN

ALIGNC
0:  MAKE_CARRY
    movl %eax,%ebx
    movl R_A,%eax
    adcb %bl,%al
    daa
    movb %al,%bl
    movb %ah,%al
    adcb %bh,%al
    daa
    movb %al,%bh
    movw %bx,R_A
    lahf
    jno 1f
    SET_FLAG(SNES_FLAG_V)
1:  STORE_FLAGS_FROM SNES_FLAG_C,%ah
    testw %bx,%bx
    STORE_FLAGS(SNES_FLAG_NZ)
    jmp CPU_RETURN
.endm

.macro EM_AND label,addr
ALIGNC
EM_AND_\label:
    READ8_\addr
    andb %al,R_A
    STORE_FLAGS (SNES_FLAG_NZ)
    jmp CPU_RETURN
.endm

.macro E0_AND label,addr
ALIGNC
E0_AND_\label:
    READ16_\addr
    andw %ax,R_A
    STORE_FLAGS (SNES_FLAG_NZ)
    jmp CPU_RETURN
.endm

.macro EM_ASL label,addr
ALIGNC
EM_ASL_\label:
    READ8_\addr
    addb %al,%al
    STORE_FLAGS (SNES_FLAG_NZC)
    SET_BYTE
    jmp CPU_RETURN
.endm

.macro E0_ASL label,addr
ALIGNC
E0_ASL_\label:
    READ16_\addr
    movl %eax,%edi
    addw %di,%di
    STORE_FLAGS (SNES_FLAG_NZC)
    movl %edi,%eax
    SET_WORD_HL
    jmp CPU_RETURN
.endm

.macro BFC label,flag
ALIGNC
ALL_\label:
    addl $2,_NativePC
    JUMP_FLAG \flag,CPU_RETURN  // flag set
    movsbl -1(_NativePC),%eax   // sign extend for addition
    addl $6,_Cycles             // branch taken
    addl %eax,_NativePC
    jmp CPU_RETURN
.endm

.macro BFS label,flag
ALIGNC
ALL_\label:
    addl $2,_NativePC
    JUMP_NOT_FLAG \flag,CPU_RETURN  // flag clear
    movsbl -1(_NativePC),%eax   // sign extend for addition
    addl $6,_Cycles             // branch taken
    addl %eax,_NativePC
    jmp CPU_RETURN
.endm

.macro EM_BIT label,addr
ALIGNC
EM_BIT_\label:
    READ8_\addr
    CLR_FLAG(SNES_FLAG_N+SNES_FLAG_V+SNES_FLAG_Z)
    testb %al,%al
    jns 1f
    SET_FLAG(SNES_FLAG_N)
1:  testb $0x40,%al
    jz  1f
    SET_FLAG(SNES_FLAG_V)
1:  testb R_A,%al
    jnz 1f
    SET_FLAG(SNES_FLAG_Z)
1:  jmp CPU_RETURN
.endm

.macro E0_BIT label,addr
ALIGNC
E0_BIT_\label:
    READ16_\addr
    CLR_FLAG(SNES_FLAG_N+SNES_FLAG_V+SNES_FLAG_Z)
    testb %ah,%ah
    jns 1f
    SET_FLAG(SNES_FLAG_N)
1:  testb $0x40,%ah
    jz  1f
    SET_FLAG(SNES_FLAG_V)
1:  testl R_A,%eax
    jnz 1f
    SET_FLAG(SNES_FLAG_Z)
1:  jmp CPU_RETURN
.endm

.macro EM_CMP label,addr
ALIGNC
EM_CMP_\label:
    READ8_\addr
    cmpb %al,R_A
    cmc
    STORE_FLAGS (SNES_FLAG_NZC)
    jmp CPU_RETURN
.endm

.macro E0_CMP label,addr
ALIGNC
E0_CMP_\label:
    READ16_\addr
    cmpw %ax,R_A
    cmc
    STORE_FLAGS (SNES_FLAG_NZC)
    jmp CPU_RETURN
.endm

.macro EX_CPX label,addr
ALIGNC
EX_CPX_\label:
    READ8_\addr
    cmpb %al,R_X
    cmc
    STORE_FLAGS (SNES_FLAG_NZC)
    jmp CPU_RETURN
.endm

.macro E0_CPX label,addr
ALIGNC
E0_CPX_\label:
    READ16_\addr
    cmpw %ax,R_X
    cmc
    STORE_FLAGS (SNES_FLAG_NZC)
    jmp CPU_RETURN
.endm

.macro EX_CPY label,addr
ALIGNC
EX_CPY_\label:
    READ8_\addr
    cmpb %al,R_Y
    cmc
    STORE_FLAGS (SNES_FLAG_NZC)
    jmp CPU_RETURN
.endm

.macro E0_CPY label,addr
ALIGNC
E0_CPY_\label:
    READ16_\addr
    cmpw %ax,R_Y
    cmc
    STORE_FLAGS (SNES_FLAG_NZC)
    jmp CPU_RETURN
.endm

.macro EM_DEC label,addr
ALIGNC
EM_DEC_\label:
    READ8_\addr
    decl %eax
    SET_BYTE
    testb %al,%al
    STORE_FLAGS (SNES_FLAG_NZ)
    jmp CPU_RETURN
.endm

.macro E0_DEC label,addr
ALIGNC
E0_DEC_\label:
    READ16_\addr
    decl %eax
    SET_WORD_HL
    testw %ax,%ax
    STORE_FLAGS (SNES_FLAG_NZ)
    jmp CPU_RETURN
.endm

.macro EM_EOR label,addr
ALIGNC
EM_EOR_\label:
    READ8_\addr
    xorb %al,R_A
    STORE_FLAGS (SNES_FLAG_NZ)
    jmp CPU_RETURN
.endm

.macro E0_EOR label,addr
ALIGNC
E0_EOR_\label:
    READ16_\addr
    xorw %ax,R_A
    STORE_FLAGS (SNES_FLAG_NZ)
    jmp CPU_RETURN
.endm

.macro EM_INC label,addr
ALIGNC
EM_INC_\label:
    READ8_\addr
    incl %eax
    SET_BYTE
    testb %al,%al
    STORE_FLAGS (SNES_FLAG_NZ)
    jmp CPU_RETURN
.endm

.macro E0_INC label,addr
ALIGNC
E0_INC_\label:
    READ16_\addr
    incl %eax
    SET_WORD_HL
    testw %ax,%ax
    STORE_FLAGS (SNES_FLAG_NZ)
    jmp CPU_RETURN
.endm

.macro EM_LDA label,addr
ALIGNC
EM_LDA_\label:
    READ8_\addr
    movb %al,R_A
    testb %al,%al
    STORE_FLAGS (SNES_FLAG_NZ)
    jmp CPU_RETURN
.endm

.macro E0_LDA label,addr
ALIGNC
E0_LDA_\label:
    READ16_\addr
    movw %ax,R_A
    testw %ax,%ax
    STORE_FLAGS (SNES_FLAG_NZ)
    jmp CPU_RETURN
.endm

.macro EX_LDX label,addr
ALIGNC
EX_LDX_\label:
    READ8_\addr
    movb %al,R_X
    testb %al,%al
    STORE_FLAGS (SNES_FLAG_NZ)
    jmp CPU_RETURN
.endm

.macro E0_LDX label,addr
ALIGNC
E0_LDX_\label:
    READ16_\addr
    movw %ax,R_X
    testw %ax,%ax
    STORE_FLAGS (SNES_FLAG_NZ)
    jmp CPU_RETURN
.endm

.macro EX_LDY label,addr
ALIGNC
EX_LDY_\label:
    READ8_\addr
    movb %al,R_Y
    testb %al,%al
    STORE_FLAGS (SNES_FLAG_NZ)
    jmp CPU_RETURN
.endm

.macro E0_LDY label,addr
ALIGNC
E0_LDY_\label:
    READ16_\addr
    movw %ax,R_Y
    testw %ax,%ax
    STORE_FLAGS (SNES_FLAG_NZ)
    jmp CPU_RETURN
.endm

.macro EM_LSR label,addr
ALIGNC
EM_LSR_\label:
    READ8_\addr
    shrb $1,%al
    STORE_FLAGS (SNES_FLAG_NZC)
    SET_BYTE
    jmp CPU_RETURN
.endm

.macro E0_LSR label,addr
ALIGNC
E0_LSR_\label:
    READ16_\addr
    movl %eax,%edi
    shrw $1,%di
    STORE_FLAGS (SNES_FLAG_NZC)
    movl %edi,%eax
    SET_WORD_HL
    jmp CPU_RETURN
.endm

.macro EM_ORA label,addr
ALIGNC
EM_ORA_\label:
    READ8_\addr
    orb %al,R_A
    STORE_FLAGS (SNES_FLAG_NZ)
    jmp CPU_RETURN
.endm

.macro E0_ORA label,addr
ALIGNC
E0_ORA_\label:
    READ16_\addr
    orw %ax,R_A
    STORE_FLAGS (SNES_FLAG_NZ)
    jmp CPU_RETURN
.endm

.macro EM_ROL label,addr
ALIGNC
EM_ROL_\label:
    READ8_\addr
    MAKE_CARRY
    adcb %al,%al
    STORE_FLAGS (SNES_FLAG_NZC)
    SET_BYTE
    jmp CPU_RETURN
.endm

.macro E0_ROL label,addr
ALIGNC
E0_ROL_\label:
    READ16_\addr
    movl %eax,%edi
    MAKE_CARRY
    adcw %di,%di
    STORE_FLAGS (SNES_FLAG_NZC)
    movl %edi,%eax
    SET_WORD_HL
    jmp CPU_RETURN
.endm

.macro EM_ROR label,addr
ALIGNC
EM_ROR_\label:
    READ8_\addr
    shrb $1,_P_B    //MAKE_CARRY
    rcrb $1,%al
    adcb _P_B,_P_B  //STORE_FLAGS (SNES_FLAG_C)
    testb %al,%al
    STORE_FLAGS (SNES_FLAG_NZ)
    SET_BYTE
    jmp CPU_RETURN
.endm

.macro E0_ROR label,addr
ALIGNC
E0_ROR_\label:
    READ16_\addr
    movl %eax,%edi
    shrb $1,_P_B    //MAKE_CARRY
    rcrw $1,%di
    adcb _P_B,_P_B  //STORE_FLAGS (SNES_FLAG_C)
    testw %di,%di
    STORE_FLAGS (SNES_FLAG_NZ)
    movl %edi,%eax
    SET_WORD_HL
    jmp CPU_RETURN
.endm

.macro EM_SBC label,addr
ALIGNC
EM_SBC_\label:
    READ8_\addr
    CLR_FLAG(SNES_FLAG_V)
    JUMP_FLAG SNES_FLAG_D,0f
    MAKE_NOT_CARRY
    sbbb %al,R_A
    cmc
    lahf
    jno 2f
    SET_FLAG(SNES_FLAG_V)
2:  STORE_FLAGS_FROM SNES_FLAG_NZC,%ah
    jmp CPU_RETURN

ALIGNC
0:  MAKE_NOT_CARRY
    movb R_A,%ah
    sbbb %al,%ah
    movb %ah,%al
    das
    movb %al,R_A
    cmc
    lahf
    jno 2f
1:  SET_FLAG(SNES_FLAG_V)
2:  STORE_FLAGS_FROM SNES_FLAG_NZC,%ah
    jmp CPU_RETURN
.endm

.macro E0_SBC label,addr
ALIGNC
E0_SBC_\label:
    READ16_\addr
    CLR_FLAG(SNES_FLAG_V)
    JUMP_FLAG SNES_FLAG_D,0f
    MAKE_NOT_CARRY
    sbbw %ax,R_A
    cmc
    lahf
    jno 1f
    SET_FLAG(SNES_FLAG_V)
1:  STORE_FLAGS_FROM SNES_FLAG_NZC,%ah
    jmp CPU_RETURN

ALIGNC
0:  MAKE_NOT_CARRY
    movl %eax,%ebx
    movl R_A,%eax
    sbbb %bl,%al
    das
    movb %al,%bl
    movb %ah,%al
    sbbb %bh,%al
    das
    movb %al,%bh
    movw %bx,R_A
    cmc
    lahf
    jno 1f
    SET_FLAG(SNES_FLAG_V)
1:  STORE_FLAGS_FROM SNES_FLAG_C,%ah
    testw %bx,%bx
    STORE_FLAGS(SNES_FLAG_NZ)
    jmp CPU_RETURN
.endm

.macro EM_STA label,addr
ALIGNC
EM_STA_\label:
    ADDR_\addr
    movl R_A,%eax
    SET_BYTE
    jmp CPU_RETURN
.endm

.macro E0_STA label,addr
ALIGNC
E0_STA_\label:
    ADDR_\addr
    movl R_A,%eax
    SET_WORD NoSave=1
    jmp CPU_RETURN
.endm

.macro EX_STX label,addr
ALIGNC
EX_STX_\label:
    ADDR_\addr
    movl R_X,%eax
    SET_BYTE
    jmp CPU_RETURN
.endm

.macro E0_STX label,addr
ALIGNC
E0_STX_\label:
    ADDR_\addr
    movl R_X,%eax
    SET_WORD NoSave=1
    jmp CPU_RETURN
.endm

.macro EX_STY label,addr
ALIGNC
EX_STY_\label:
    ADDR_\addr
    movl R_Y,%eax
    SET_BYTE
    jmp CPU_RETURN
.endm

.macro E0_STY label,addr
ALIGNC
E0_STY_\label:
    ADDR_\addr
    movl R_Y,%eax
    SET_WORD NoSave=1
    jmp CPU_RETURN
.endm

.macro EM_STZ label,addr
ALIGNC
EM_STZ_\label:
    ADDR_\addr
    xorl %eax,%eax
    SET_BYTE
    jmp CPU_RETURN
.endm

.macro E0_STZ label,addr
ALIGNC
E0_STZ_\label:
    ADDR_\addr
    xorl %eax,%eax
    SET_WORD Double=1
    jmp CPU_RETURN
.endm

// TRB set flag Z on A & Memory Operand not A & !Memory Operand!
.macro EM_TRB label,addr
ALIGNC
EM_TRB_\label:
    READ8_\addr
    testb %al,R_A
    STORE_FLAGS (SNES_FLAG_Z)
    movb R_A,%ah
    xorb $-1,%ah
    andb %ah,%al
    SET_BYTE
    jmp CPU_RETURN
.endm

.macro E0_TRB label,addr
ALIGNC
E0_TRB_\label:
    READ16_\addr
    movl %eax,%edi
    testl R_A,%edi
    STORE_FLAGS (SNES_FLAG_Z)
    movl R_A,%eax
    xorl $-1,%eax
    andl %edi,%eax
    SET_WORD_HL
    jmp CPU_RETURN
.endm

// TSB set flag Z on A & Memory Operand not A | Memory Operand!

.macro EM_TSB label,addr
ALIGNC
EM_TSB_\label:
    READ8_\addr
    testb %al,R_A
    STORE_FLAGS (SNES_FLAG_Z)
    orb R_A,%al
    SET_BYTE
    jmp CPU_RETURN
.endm

.macro E0_TSB label,addr
ALIGNC
E0_TSB_\label:
    READ16_\addr
    movl %eax,%edi
    testl R_A,%edi
    STORE_FLAGS (SNES_FLAG_Z)
    movl R_A,%eax
    orl %edi,%eax
    SET_WORD_HL
    jmp CPU_RETURN
.endm
