/*

 DMA.S - (H)DMA emulation

*/

#include "misc.i"
#include "DMA.i"
#include "regs.i"

EXPORT_C DMA_start

.data
ALIGND
DMA_Data_Areas:
.long TableDMA0,TableDMA1,TableDMA2,TableDMA3
.long TableDMA4,TableDMA5,TableDMA6,TableDMA7

.globl DMA_PPU_Order
DMA_PPU_Order:
.byte   0,0,0,0
.byte   0,1,0,1
.byte   0,0,0,0
.byte   0,0,1,1
.byte   0,1,2,3
.byte   0,1,0,1
.byte   0,0,0,0
.byte   0,0,1,1

.globl HDMA_Size
HDMA_Size:
.byte   1,2,2,4
.byte   4,2,2,4

.bss
ALIGND
.macro DMA_DATA num
ALIGND
.globl TableDMA\num
TableDMA\num:
.globl DMAP_\num,BBAD_\num,DMA_Vid_\num,NTRL_\num,HDMA_On_\num
// dh0bcttt d=CPU->PPU,h=addr,b=inc/dec,c=inc/fixed,t=type
DMAP_\num:      skipb
BBAD_\num:      // Low byte of 0x21??
DMA_Vid_\num:   skipb
NTRL_\num:      skipb
DMA_Inc_\num:   skipb
.globl A1T_\num,A1TL_\num,A1TH_\num,A1B_\num
A1T_\num:               // Source Address L/H/B
A1TL_\num:      skipb   // Source Address L
A1TH_\num:      skipb   // Source Address H
A1B_\num:       skipb   // Source Bank Address
                skipb
.globl DAS_\num,DASL_\num,DASH_\num,DASB_\num
DAS_\num:               // Data Size L/H
DASL_\num:      skipb   // Data Size L
DASH_\num:      skipb   // Data Size H
DASB_\num:      skipb   // Data address bank
                skipb
.globl A2T_\num,A2L_\num,A2H_\num,A2B_\num
A2T_\num:
A2L_\num:       skipb   // HDMA table address L
A2H_\num:       skipb   // HDMA table address H
A2B_\num:       skipb   // HDMA table bank address
                skipb

// DMA_Wr#_x - These hold the write handlers for DMA
// DMA_Rd#_x - These hold the read handlers for DMA
// HDMA_Siz_x - These hold the register size for HDMA
.globl DMA_Wr0_\num,DMA_Wr1_\num,DMA_Wr2_\num,DMA_Wr3_\num
.globl DMA_Rd0_\num,DMA_Rd1_\num,DMA_Rd2_\num,DMA_Rd3_\num
.globl HDMA_Siz_\num
ALIGND
DMA_Wr0_\num:   skipl
DMA_Wr1_\num:   skipl
DMA_Wr2_\num:   skipl
DMA_Wr3_\num:   skipl
DMA_Rd0_\num:   skipl
DMA_Rd1_\num:   skipl
DMA_Rd2_\num:   skipl
DMA_Rd3_\num:   skipl
ABank_Rd_\num:  skipl
ABank_Wr_\num:  skipl
IBank_Rd_\num:  skipl
HDMA_Siz_\num:  skipl
.endm

DMA_DATA 0
DMA_DATA 1
DMA_DATA 2
DMA_DATA 3
DMA_DATA 4
DMA_DATA 5
DMA_DATA 6
DMA_DATA 7

.globl MDMAEN
MDMAEN: skipb   // DMA enable
.globl HDMAEN
HDMAEN: skipb   // HDMA enable
.globl HDMAON
HDMAON: skipb   // HDMA enabled this refresh

.text
.macro call2 retaddx,call2,call1
  movl \call2,%eax
  pushl \retaddx
  pushl %eax    // omit the write handler call
  jmpl *\call1
.endm

ALIGNC
/*
  al = data byte
  ebx = A-bus address
  ebp = DAS (byte count)
  esi = Address adjustment
 */
Do_DMA_Channel:
  movl A1T(%edx),%ebx   // CPU address in ebx

  xorl %ebp,%ebp
  movw DAS(%edx),%bp
  decl %ebp
  andl $0xFFFF,%ebp
  incl %ebp
  
0:movb DMAP(%edx),%al
  movl $0x2100,%ecx
  movsbl DMA_Inc(%edx),%esi // Get address adjustment
  testb %al,%al     // Is the operation CPU->PPU?
  jns 0f

// PPU->CPU
1:call2 $5f,ABank_Wr(%edx),DMA_Rd0(%edx)
5:addw %si,%bx
  decl %ebp
  jz  1f

  call2 $5f,ABank_Wr(%edx),DMA_Rd1(%edx)
5:addw %si,%bx
  decl %ebp
  jz  1f

  call2 $5f,ABank_Wr(%edx),DMA_Rd2(%edx)
5:addw %si,%bx
  decl %ebp
  jz  1f

  call2 $5f,ABank_Wr(%edx),DMA_Rd3(%edx)
5:addw %si,%bx
  decl %ebp
  jnz 1b
1:movw %bx,A1T(%edx)     // v0.15 forgot to update DMA pointers!
  ret

ALIGNC
// CPU->PPU
0:call2 $5f,DMA_Wr0(%edx),ABank_Rd(%edx)
5:addw %si,%bx
  decl %ebp
  jz  1f

  call2 $5f,DMA_Wr1(%edx),ABank_Rd(%edx)
5:addw %si,%bx
  decl %ebp
  jz  1f

  call2 $5f,DMA_Wr2(%edx),ABank_Rd(%edx)
5:addw %si,%bx
  decl %ebp
  jz  1f

  call2 $5f,DMA_Wr3(%edx),ABank_Rd(%edx)
5:addw %si,%bx
  decl %ebp
  jnz 0b

1:movw %bx,A1T(%edx)     // v0.15 forgot to update DMA pointers!
  movw $0,DAS(%edx)
  ret

ALIGNC
.globl Do_HDMA_Channel
Do_HDMA_Channel:
  movl A2T(%edx),%ebx     // Get table address
  movb DMAP(%edx),%al     // Get HDMA control byte
  testb $0x40,%al         // Check for indirect addressing
  movl HDMA_Siz(%edx),%ecx  // Get HDMA transfer size
  jnz Do_HDMA_Indirect

Do_HDMA_Absolute:
  movb NTRL(%edx),%ah       // Get number of lines to transfer
  testb $0x7F,%ah           // Need new set?
  jz  HDMA_Absolute_Next_Set
  testb %ah,%ah
  js  2f
  jmp HDMA_Absolute_Continue

HDMA_Absolute_Next_Set:
  call ABank_Rd(%edx)
  incw %bx                // Adjust table address
  testb %al,%al           // Check for zero-length set
  movb %al,NTRL(%edx)     // Save length of set
  jz  0f
  movl %ebx,A2T(%edx)     // Save new table address
2:
  call2 $5f,DMA_Wr0(%edx),ABank_Rd(%edx)
5:cmpb $2,%cl
  incw %bx                // Adjust temporary table pointer
  jb  1f

  call2 $5f,DMA_Wr1(%edx),ABank_Rd(%edx)
5:cmpb $4,%cl
  incw %bx                // Adjust temporary table pointer
  jb  1f

  call2 $5f,DMA_Wr2(%edx),ABank_Rd(%edx)
5:incw %bx

  call2 $1f,DMA_Wr3(%edx),ABank_Rd(%edx)
1:addw %cx,A2T(%edx)
HDMA_Absolute_Continue:
  decb NTRL(%edx)
  stc
  ret

0:movl %ebx,A2T(%edx)
  clc
  ret

Do_HDMA_Indirect:
  movb NTRL(%edx),%ah     // Get number of lines to transfer
  testb $0x7F,%ah
  jz  HDMA_Indirect_Next_Set
  testb %ah,%ah
  js  2f
  jmp HDMA_Indirect_Continue

HDMA_Indirect_Next_Set:
  call ABank_Rd(%edx)
  incw %bx
  movb %al,NTRL(%edx)
  testb %al,%al
  movb %al,%ah
  jz  0b
  call ABank_Rd(%edx)
  incw %bx
  movb %al,DASL(%edx)
  call ABank_Rd(%edx)
  incw %bx
  movb %al,DASH(%edx)
  movl %ebx,A2T(%edx)
2:movl DAS(%edx),%ebx

  call2 $5f,DMA_Wr0(%edx),IBank_Rd(%edx)
5:cmpb $2,%cl
  incw %bx                // Adjust temporary table pointer
  jb  1f

  call2 $5f,DMA_Wr1(%edx),IBank_Rd(%edx)
5:cmpb $4,%cl
  incw %bx                // Adjust temporary table pointer
  jb  1f

  call2 $5f,DMA_Wr2(%edx),IBank_Rd(%edx)
5:incw %bx

  call2 $1f,DMA_Wr3(%edx),IBank_Rd(%edx)
1:addw %cx,DAS(%edx)
HDMA_Indirect_Continue:
  decb NTRL(%edx)
  stc
  ret

.macro DMAOPERATION num
  movb MDMAEN,%al
  testb $(1<<\num),%al
  jz  7f

  LOAD_DMA_TABLE \num
  call Do_DMA_Channel
7:
.endm

ALIGNC
.globl SNES_R420B
SNES_R420B: // MDMAEN
  movb $0,%al
  ret

ALIGNC
.globl SNES_R420C
SNES_R420C: // HDMAEN
  movb HDMAON,%al
  ret

ALIGNC
.globl SNES_W420B
SNES_W420B: // MDMAEN
  movb %al,MDMAEN
  pusha

  movb CPU_Executing,%al
  pushl %eax
  movb $0,CPU_Executing

  DMAOPERATION 0
  DMAOPERATION 1
  DMAOPERATION 2
  DMAOPERATION 3
  DMAOPERATION 4
  DMAOPERATION 5
  DMAOPERATION 6
  DMAOPERATION 7

  popl %eax
  movb %al,CPU_Executing

  popa
8:ret

ALIGNC
.globl SNES_W420C
SNES_W420C: // HDMAEN   // Actually handled within screen core!
  movb %al,HDMAEN
//ret
  testb $0x80,HVBJOY    // Vblank?
  jnz 1f
  cmpl $1024-40,_65c816_Cycles  // Hblank?
  jnb 1f
  testb $0x80,_INIDISP  // Force-blank?
  jns 0f
1:
/*pusha
  RELATCH_HDMA_IN_FRAME 0
  RELATCH_HDMA_IN_FRAME 1
  RELATCH_HDMA_IN_FRAME 2
  RELATCH_HDMA_IN_FRAME 3
  RELATCH_HDMA_IN_FRAME 4
  RELATCH_HDMA_IN_FRAME 5
  RELATCH_HDMA_IN_FRAME 6
  RELATCH_HDMA_IN_FRAME 7
  popa*/
  movb %al,HDMAON
0:ret

ALIGNC
.globl do_HDMA
do_HDMA:
  movb HDMAON,%al
  testb %al,%al
  jz  9f
  pusha
  HDMAOPERATION 0
  HDMAOPERATION 1
  HDMAOPERATION 2
  HDMAOPERATION 3
  HDMAOPERATION 4
  HDMAOPERATION 5
  HDMAOPERATION 6
  HDMAOPERATION 7
  popa
9:ret

// Requires %eax to be 0x00FFFFFF!
.macro Reset_DMA_Channel num
  movb %al,DMAP_\num
  movb %al,BBAD_\num
  movb %al,NTRL_\num
  movl %eax,A1T_\num
  movl %eax,DAS_\num
  movl %eax,A2T_\num

  pushl %ebx
  pushl %ecx
  movl $UNSUPPORTED_READ,%ebx
  movl $UNSUPPORTED_WRITE,%ecx
  movl %ebx,DMA_Rd0_\num
  movl %ecx,DMA_Wr0_\num
  movl %ebx,DMA_Rd1_\num
  movl %ecx,DMA_Wr1_\num
  movl %ebx,DMA_Rd2_\num
  movl %ecx,DMA_Wr2_\num
  movl %ebx,DMA_Rd3_\num
  movl %ecx,DMA_Wr3_\num
  movl %ebx,ABank_Rd_\num
  movl %ecx,ABank_Wr_\num
  popl %ecx
  popl %ebx

  movb $0,DMA_Inc_\num
.endm

.globl Reset_DMA
Reset_DMA:
  // Set eax to 0...
  xorl %eax,%eax
  movb %al,MDMAEN
  movb %al,HDMAEN
  movb %al,HDMAON

  // Now 0x00FFFFFF...
  movl $0x00FFFFFF,%eax
  Reset_DMA_Channel 0
  Reset_DMA_Channel 1
  Reset_DMA_Channel 2
  Reset_DMA_Channel 3
  Reset_DMA_Channel 4
  Reset_DMA_Channel 5
  Reset_DMA_Channel 6
  Reset_DMA_Channel 7

  // Back to 0...
  xorl %eax,%eax

  ret

ALIGNC
Update_DMA_PPU_Handlers:
  shrl $2,%edi
  pushl %eax
  andl $7 << 2,%edi
  pushl %ebx
  pushl %ecx
  pushl %edx
  movl DMA_Data_Areas(%edi),%ebx
  xorl %edx,%edx
  movb DMAP(%ebx),%al
  movb DMA_Vid(%ebx),%dl    // PPU address in edx
  andl $7,%eax
  movb HDMA_Size(%eax),%cl
  leal DMA_PPU_Order(,%eax,4),%eax
  movb %cl,HDMA_Siz(%ebx)
  movl %edx,%ecx            // PPU address in ecx
//addb (%eax),%dl
  movl Read_Map_21(,%edx,4),%edi
  movl Write_Map_21(,%edx,4),%edx
  movl %edi,DMA_Rd0(%ebx)
  movl %edx,DMA_Wr0(%ebx)
  movl %ecx,%edx            // PPU address in ecx
  addb 1(%eax),%cl
  movl Read_Map_21(,%ecx,4),%edi
  movl Write_Map_21(,%ecx,4),%ecx
  movl %edi,DMA_Rd1(%ebx)
  movl %ecx,DMA_Wr1(%ebx)
  movl %edx,%ecx            // PPU address in edx
  addb 2(%eax),%dl
  movl Read_Map_21(,%edx,4),%edi
  movl Write_Map_21(,%edx,4),%edx
  movl %edi,DMA_Rd2(%ebx)
  movl %edx,DMA_Wr2(%ebx)
  addb 3(%eax),%cl
  movl Read_Map_21(,%ecx,4),%edi
  movl Write_Map_21(,%ecx,4),%ecx
  movl %edi,DMA_Rd3(%ebx)
  movl %ecx,DMA_Wr3(%ebx)
  popl %edx
  popl %ecx
  popl %ebx
  popl %eax
  ret

// Read from 43xx handlers
.macro MAP_READ_DMA num
ALIGNC
.globl MAP_READ_DMAP\num
MAP_READ_DMAP\num:
  movb DMAP_\num,%al
  ret

ALIGNC
.globl MAP_READ_BBAD\num
MAP_READ_BBAD\num:
  movb BBAD_\num,%al
  ret

ALIGNC
.globl MAP_READ_A1TL\num
MAP_READ_A1TL\num:
  movb A1TL_\num,%al
  ret

ALIGNC
.globl MAP_READ_A1TH\num
MAP_READ_A1TH\num:
  movb A1TH_\num,%al
  ret

ALIGNC
.globl MAP_READ_A1B\num
MAP_READ_A1B\num:
  movb A1B_\num,%al
  ret

ALIGNC
.globl MAP_READ_DASL\num
MAP_READ_DASL\num:
  movb DASL_\num,%al
  ret

ALIGNC
.globl MAP_READ_DASH\num
MAP_READ_DASH\num:
  movb DASH_\num,%al
  ret

ALIGNC
.globl MAP_READ_DASB\num
MAP_READ_DASB\num:
  movb DASB_\num,%al
  ret

ALIGNC
.globl MAP_READ_A2L\num
MAP_READ_A2L\num:
  movb A2L_\num,%al
  ret

ALIGNC
.globl MAP_READ_A2H\num
MAP_READ_A2H\num:
  movb A2H_\num,%al
  ret

ALIGNC
.globl MAP_READ_NTRL\num
MAP_READ_NTRL\num:
  movb NTRL_\num,%al
  ret
.endm

MAP_READ_DMA 0
MAP_READ_DMA 1
MAP_READ_DMA 2
MAP_READ_DMA 3
MAP_READ_DMA 4
MAP_READ_DMA 5
MAP_READ_DMA 6
MAP_READ_DMA 7

// Write to 43xx handlers
.macro MAP_WRITE_DMA num
ALIGNC
.globl MAP_WRITE_DMAP\num
MAP_WRITE_DMAP\num:
  cmpb %al,DMAP_\num
  je 0f

  pushl %ebx
  movb %al,DMAP_\num

  testb $8,%al      // Does the operation require address adjustment?
  movb $0,%bl
  jnz 1f

  decl %ebx         // Set address decrement
  testb $0x10,%al
  jnz 1f

  addb $2,%bl       // Set address increment
1:movb %bl,DMA_Inc_\num

  popl %ebx
  jmp Update_DMA_PPU_Handlers   // It'll return for us
0:ret

ALIGNC
.globl MAP_WRITE_BBAD\num
MAP_WRITE_BBAD\num:
  cmpb %al,DMAP_\num
  je 0f
  movb %al,BBAD_\num
  jmp Update_DMA_PPU_Handlers   // It'll return for us
0:ret

ALIGNC
.globl MAP_WRITE_A1TL\num
MAP_WRITE_A1TL\num:
  movb %al,A1TL_\num
  ret

ALIGNC
.globl MAP_WRITE_A1TH\num
MAP_WRITE_A1TH\num:
  movb %al,A1TH_\num
  ret

ALIGNC
.globl MAP_WRITE_A1B\num
MAP_WRITE_A1B\num:
  cmpb %al,A1B_\num
  je 0f
  pushl %ebx
  xorl %ebx,%ebx
  pushl %eax
  movb %al,%bl
  movb %al,A1B_\num
  movb %al,A2B_\num
  movl _Read_BankMapping(,%ebx,4),%eax
  movl _Write_BankMapping(,%ebx,4),%ebx
  movl %eax,ABank_Rd_\num
  movl %ebx,ABank_Wr_\num
  popl %eax
  popl %ebx
0:ret

ALIGNC
.globl MAP_WRITE_DASL\num
MAP_WRITE_DASL\num:
  movb %al,DASL_\num
  ret

ALIGNC
.globl MAP_WRITE_DASH\num
MAP_WRITE_DASH\num:
  movb %al,DASH_\num
  ret

ALIGNC
.globl MAP_WRITE_DASB\num
MAP_WRITE_DASB\num:
  cmpb %al,DASB_\num
  je 0f
  pushl %ebx
  xorl %ebx,%ebx
  movb %al,%bl
  movl _Read_BankMapping(,%ebx,4),%ebx
  movb %al,DASB_\num
  movl %ebx,IBank_Rd_\num
  popl %ebx
0:ret

ALIGNC
.globl MAP_WRITE_A2L\num
MAP_WRITE_A2L\num:
  movb %al,A2L_\num
  ret

ALIGNC
.globl MAP_WRITE_A2H\num
MAP_WRITE_A2H\num:
  movb %al,A2H_\num
  ret

ALIGNC
.globl MAP_WRITE_NTRL\num
MAP_WRITE_NTRL\num:
  movb %al,NTRL_\num
  ret
.endm

MAP_WRITE_DMA 0
MAP_WRITE_DMA 1
MAP_WRITE_DMA 2
MAP_WRITE_DMA 3
MAP_WRITE_DMA 4
MAP_WRITE_DMA 5
MAP_WRITE_DMA 6
MAP_WRITE_DMA 7

.text; ALIGNC
.data; ALIGND
.bss; ALIGND
