;%define TRAP_BAD_READS
;%define TRAP_BAD_WRITES
;%define TRAP_IGNORED_WRITES
;
;
; SNEeSe SNES memory mapper v1.6 - NASM rewrite
;
; There are numerous jump tables in this code, but a lot less than before...
;  back down from 4k to 256... easier on the cache this way, and with
;  dynamic memory map init both LoROM and HiROM are very simply mapped...
;
;  Files are split up as follows:
;
;    mem.S      - LoROM memory map code/tables.
;    memhi.S    - HiROM memory map code/tables.
;    memmap.i   - Memory mapping macros.
;    memblank.S - 'Blank' ROM bank handlers for dynamic memory map
;                 and code common to both LoROM and HiROM.
;    memhard.S  - Hardware port (21xx/40xx/42xx/43xx) handlers.
;    memDMA.S   - DMA emulation code.
;
;

%include "misc.ni"
%include "memmap.ni"

EXPORT_C memblank_start

extern _SRAM_READ
extern _RomReadAdjust,_RomHiReadAdjust
extern _WRAM,_SRAM_Mask,_SRAM
extern Read_Map_20_5F,Write_Map_20_5F

section .data
global _LoROM_Blank
ALIGND
_LoROM_Blank:
; Banks 00-3F are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF RESERVED 8000-FFFF ROM
    DUPLICATE dd,64,_BlankL_00_3F

; Banks 40-5F are 0000-7FFF RESERVED 8000-FFFF ROM
    DUPLICATE dd,32,_BlankL_40_5F

; Banks 60-6F are not mapped!
    DUPLICATE dd,16,UNSUPPORTED_READ

; Banks 70-73 is SRAM (mirrored)
    DUPLICATE dd,4,_SRAM_READ

; Banks 74-7D are not mapped!
    DUPLICATE dd,10,UNSUPPORTED_READ

; Bank  7E-7F is  0000-FFFF RAM Nb 7E is used to store that 8KB block thats repeated in mem
    DUPLICATE dd,2,_RAM_READ

; Banks 80-BF are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF RESERVED 8000-FFFF ROM
    DUPLICATE dd,64,_BlankL_80_BF

; Banks C0-DF are 0000-7FFF RESERVED 8000-FFFF ROM
    DUPLICATE dd,64,_BlankL_C0_FF

global _HiROM_Blank
ALIGND
_HiROM_Blank:
; Banks 00-2F are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF RESERVED 8000-FFFF ROM
; Banks 30-3F are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF SRAM 8000-FFFF ROM
    DUPLICATE dd,64,_BlankH_00_3F_No_SRAM

; Banks 40-5F are 0000-FFFF ROM
    DUPLICATE dd,32,_BlankH_40_5F

; Banks 60-7D are not mapped!
    DUPLICATE dd,30,UNSUPPORTED_READ

; Bank  7E-7F is  0000-FFFF RAM Nb 7E is used to store that 8KB block thats repeated in mem
    DUPLICATE dd,2,_RAM_READ

; Banks 80-AF are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF RESERVED 8000-FFFF ROM
; Banks B0-BF are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF SRAM 8000-FFFF ROM
    DUPLICATE dd,64,_BlankH_80_BF_No_SRAM

; Banks C0-FF are 0000-FFFF ROM
    DUPLICATE dd,64,_BlankH_C0_FF

section .bss
ALIGND
global _Read_BankMapping
_Read_BankMapping:  skipl 256

ALIGND
global _Write_BankMapping
_Write_BankMapping: skipl 256

ALIGND
global _Read_BankOffset
_Read_BankOffset:   skipl 256*2

extern _Map_Address
extern _Map_Byte

section .text
global UNSUPPORTED_READ,_UNSUPPORTED_READ
ALIGNC
_UNSUPPORTED_READ:
UNSUPPORTED_READ:
    mov al,0
%ifdef DEBUG
%ifdef TRAP_BAD_READS
extern _InvalidHWRead
    mov [_Map_Address],ebx  ; Set up Map Address so message works!
    mov [_Map_Byte],al      ; Set up Map Byte so message works
    pusha
    call _InvalidHWRead     ; Unmapped hardware address!
    popa
%endif
%endif
    ret

global UNSUPPORTED_WRITE,_UNSUPPORTED_WRITE
ALIGNC
_UNSUPPORTED_WRITE:
UNSUPPORTED_WRITE:
%ifdef DEBUG
%ifdef TRAP_BAD_WRITES
extern _InvalidHWWrite
    mov [_Map_Address],ebx  ; Set up Map Address so message works!
    mov [_Map_Byte],al      ; Set up Map Byte so message works
    pusha
    call _InvalidHWWrite    ; Unmapped hardware address!
    popa
%endif
%endif
    ret

global IGNORE_WRITE,_IGNORE_WRITE
ALIGNC
_IGNORE_WRITE:
IGNORE_WRITE:
%ifdef DEBUG
%ifdef TRAP_IGNORED_WRITES
extern _InvalidHWWrite
    mov [_Map_Address],ebx  ; Set up Map Address so message works!
    mov [_Map_Byte],al      ; Set up Map Byte so message works
    pusha
    call _InvalidHWWrite    ; Unmapped hardware address!
    popa
%endif
%endif
    ret

global _RAM_READ
ALIGNC
_RAM_READ:
    mov al,[_WRAM-0x7E0000+ebx]
    ret

global _RAM_WRITE
ALIGNC
_RAM_WRITE:
    mov [_WRAM-0x7E0000+ebx],al
    ret

global _READ_00_3F_No_SRAM
ALIGNC
_READ_00_3F_No_SRAM:
    test bh,bh
    jns short .not_rom
    ROM00_READ_32k
ALIGNC
.not_rom:
    cmp bh,0x20
    jnb short .not_lowram
    LOW_RAM_READ
ALIGNC
.not_lowram:
    cmp bh,0x60
    jnb short .not_ppu
    PPU_READ
ALIGNC
.not_ppu:
    NON_MAPPED_READ

global _READ_80_BF_No_SRAM
ALIGNC
_READ_80_BF_No_SRAM:
    test bh,bh
    jns short .not_rom
    ROM80_READ_32k
ALIGNC
.not_rom:
    cmp bh,0x20
    jnb short .not_lowram
    LOW_RAM_READ
ALIGNC
.not_lowram:
    cmp bh,0x60
    jnb short .not_ppu
    PPU_READ
ALIGNC
.not_ppu:
    NON_MAPPED_READ

global _BlankL_00_3F,_BlankL_80_BF,_BlankH_00_3F_No_SRAM,_BlankH_80_BF_No_SRAM
ALIGNC
_BlankL_00_3F:
_BlankL_80_BF:
_BlankH_00_3F_No_SRAM:
_BlankH_80_BF_No_SRAM:
    cmp bh,0x20
    jnb short .not_lowram
    LOW_RAM_READ
ALIGNC
.not_lowram:
    cmp bh,0x60
    jnb short .not_ppu
    PPU_READ
ALIGNC
.not_ppu:
    test bh,bh
    jns short .not_rom
    NO_ROM_READ
ALIGNC
.not_rom:
    NON_MAPPED_READ

ALIGNC
_BlankL_40_5F:
_BlankL_C0_FF:
    test bh,bh
    jns short .not_rom
    NO_ROM_READ
ALIGNC
.not_rom:
    NON_MAPPED_READ

%macro _BlankH_30_3F 1  ;bank
ALIGNC
global _BlankH_%1
_BlankH_%1:
    cmp bh,0x20
    jnb short .not_lowram
    LOW_RAM_READ
ALIGNC
.not_lowram:
    cmp bh,0x60
    jnb short .not_ppu
    PPU_READ
ALIGNC
.not_ppu:
    test bh,bh
    js  short .not_sram
    SRAM30_READ %1
ALIGNC
.not_sram:
    NO_ROM_READ
%endmacro

%macro _BlankH_B0_BF 1  ;bank
ALIGNC
global _BlankH_%1
_BlankH_%1:
    cmp bh,0x20
    jnb short .not_lowram
    LOW_RAM_READ
ALIGNC
.not_lowram:
    cmp bh,0x60
    jnb short .not_ppu
    PPU_READ
ALIGNC
.not_ppu:
    test bh,bh
    js  short .not_sram
    SRAMB0_READ %1
ALIGNC
.not_sram:
    NO_ROM_READ
%endmacro

ALIGNC
global _BlankH_40_5F,_BlankH_C0_FF
_BlankH_40_5F:
_BlankH_C0_FF:
    NO_ROM_READ

    _BlankH_30_3F 30
    _BlankH_30_3F 31
    _BlankH_30_3F 32
    _BlankH_30_3F 33
    _BlankH_30_3F 34
    _BlankH_30_3F 35
    _BlankH_30_3F 36
    _BlankH_30_3F 37
    _BlankH_B0_BF B0
    _BlankH_B0_BF B1
    _BlankH_B0_BF B2
    _BlankH_B0_BF B3
    _BlankH_B0_BF B4
    _BlankH_B0_BF B5
    _BlankH_B0_BF B6
    _BlankH_B0_BF B7

global _WRITE_00_3F,_WRITE_80_BF,_HWRITE_00_3F_No_SRAM,_HWRITE_80_BF_No_SRAM
ALIGNC
_WRITE_00_3F:           ; Used in LoROM!
_WRITE_80_BF:
_HWRITE_00_3F_No_SRAM:  ; Used in HiROM!
_HWRITE_80_BF_No_SRAM:
    cmp bh,0x20
    jnb short .not_lowram
    LOW_RAM_WRITE
ALIGNC
.not_lowram:
    cmp bh,0x60
    jnb short .not_ppu
    PPU_WRITE
ALIGNC
.not_ppu:
    ROM_WRITE

section .text; ALIGNC
section .data; ALIGND
section .bss; ALIGND
