;
;
; SNEeSe SNES memory mapper v1.6 - NASM rewrite
;
; Notes: HiROM has max ROM address space of 4MBytes.
;
; HiROM version
;  00-3F:0000-1FFF = Low RAM 8k (7E:0000-1FFF mirrored)
;  00-3F:2000-5FFF = Hardware ports (PPU, FX, DSP)
;  00-2F:6000-7FFF = Unmapped
;  30-3F:6000-7FFF = SRAM (up to 8k/bank) (continuously mirrored: see below)
;  00-3F:8000-FFFF = ROM 32k (C0-FF:8000-FFFF mirrored)
;  40-5F:0000-FFFF = ROM 64k (0-16Mbit)
;  60-7D:0000-FFFF = Unmapped
;  7E-7F:0000-FFFF = WRAM 128k
;  80-DF:0000-FFFF = 00-5F mirrored
;  E0-FF:0000-FFFF = ROM 64k (16-32Mbit)
;
;  SRAM has 128k (kbytes, not kbits) of address space. This is mapped
;   with 'continuous mirroring', as such:
;    2k (16kbit): 30:6000-67FF == 30:6800-6FFF == 3F:7800-7FFF
;    8k (64kbit): 30:6000-7FFF == 31:6000-7FFF == 3F:6000-7FFF
;   etc.
;  Since I know of no ROMs that require >64k of SRAM, only 64k is
;   supported (however, 128k mirroring is performed).
;

%include "misc.ni"
%include "memmap.ni"

EXPORT_C memhi_start

extern UNSUPPORTED_READ,UNSUPPORTED_WRITE
extern _RAM_READ,_RAM_WRITE
extern _READ_00_3F_No_SRAM,_READ_80_BF_No_SRAM
extern _HWRITE_00_3F_No_SRAM,_HWRITE_80_BF_No_SRAM
extern _BlankH_30,_BlankH_31,_BlankH_32,_BlankH_33
extern _BlankH_34,_BlankH_35,_BlankH_36,_BlankH_37
extern _BlankH_B0,_BlankH_B1,_BlankH_B2,_BlankH_B3
extern _BlankH_B4,_BlankH_B5,_BlankH_B6,_BlankH_B7
extern _BlankH_00_3F_No_SRAM,_BlankH_80_BF_No_SRAM
extern _HRomLinReadAdjust,_HRomHiLinReadAdjust
extern _RomReadAdjust,_RomHiReadAdjust
extern _WRAM,_SRAM_Mask,_SRAM
extern Read_Map_20_5F,Write_Map_20_5F

section .data
global _HiROM_Read
ALIGND
_HiROM_Read:

; Banks 00-2F are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF RESERVED 8000-FFFF ROM
; Banks 30-3F are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF SRAM 8000-FFFF ROM
    DUPLICATE dd,64,_READ_00_3F_No_SRAM

; Banks 40-5F are 0000-FFFF ROM
    DUPLICATE dd,32,_HREAD_40_5F

; Banks 60-7D are not mapped!
    DUPLICATE dd,30,UNSUPPORTED_READ

; Bank  7e-7f is  0000-ffff RAM Nb 7e is used to store that 8KB block thats repeated in mem
    DUPLICATE dd,2,_RAM_READ

; Banks 80-AF are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF RESERVED 8000-FFFF ROM
; Banks B0-BF are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF SRAM 8000-FFFF ROM
    DUPLICATE dd,64,_READ_80_BF_No_SRAM

; Banks C0-FF are 0000-FFFF ROM
    DUPLICATE dd,64,_HREAD_C0_FF

global _HiROM_Write
ALIGND
_HiROM_Write:

; Banks 00-2F are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF RESERVED 8000-FFFF ROM
; Banks 30-3F are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF SRAM 8000-FFFF ROM
    DUPLICATE dd,64,_HWRITE_00_3F_No_SRAM

; Banks 40-5F are 0000-FFFF ROM
    DUPLICATE dd,32,UNSUPPORTED_WRITE

; Banks 60-7D are not mapped!
    DUPLICATE dd,30,UNSUPPORTED_WRITE

; Bank  7e-7f is  0000-ffff RAM Nb 7e is used to store that 8KB block thats repeated in mem
    DUPLICATE dd,2,_RAM_WRITE

; Banks 80-AF are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF RESERVED 8000-FFFF ROM
; Banks B0-BF are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF SRAM 8000-FFFF ROM
    DUPLICATE dd,64,_HWRITE_80_BF_No_SRAM

; Banks C0-FF are 0000-FFFF ROM
    DUPLICATE dd,64,UNSUPPORTED_WRITE

global _HiROM_Read_SRAM_ROM_30,_HiROM_Read_No_SRAM_ROM_30
global _HiROM_Read_SRAM_ROM_B0,_HiROM_Read_No_SRAM_ROM_B0
global _HiROM_Read_SRAM_Blank_30,_HiROM_Read_No_SRAM_Blank_30
global _HiROM_Read_SRAM_Blank_B0,_HiROM_Read_No_SRAM_Blank_B0
global _HiROM_Write_SRAM_30,_HiROM_Write_No_SRAM_30
global _HiROM_Write_SRAM_B0,_HiROM_Write_No_SRAM_B0
ALIGND
_HiROM_Read_SRAM_ROM_30:
    dd _HREAD_30
    dd _HREAD_31
    dd _HREAD_32
    dd _HREAD_33
    dd _HREAD_34
    dd _HREAD_35
    dd _HREAD_36
    dd _HREAD_37
_HiROM_Read_SRAM_ROM_B0:
    dd _HREAD_B0
    dd _HREAD_B1
    dd _HREAD_B2
    dd _HREAD_B3
    dd _HREAD_B4
    dd _HREAD_B5
    dd _HREAD_B6
    dd _HREAD_B7
_HiROM_Read_SRAM_Blank_30:
    dd _BlankH_30
    dd _BlankH_31
    dd _BlankH_32
    dd _BlankH_33
    dd _BlankH_34
    dd _BlankH_35
    dd _BlankH_36
    dd _BlankH_37
_HiROM_Read_SRAM_Blank_B0:
    dd _BlankH_B0
    dd _BlankH_B1
    dd _BlankH_B2
    dd _BlankH_B3
    dd _BlankH_B4
    dd _BlankH_B5
    dd _BlankH_B6
    dd _BlankH_B7
_HiROM_Write_SRAM_30:
    dd _HWRITE_30
    dd _HWRITE_31
    dd _HWRITE_32
    dd _HWRITE_33
    dd _HWRITE_34
    dd _HWRITE_35
    dd _HWRITE_36
    dd _HWRITE_37
_HiROM_Write_SRAM_B0:
    dd _HWRITE_B0
    dd _HWRITE_B1
    dd _HWRITE_B2
    dd _HWRITE_B3
    dd _HWRITE_B4
    dd _HWRITE_B5
    dd _HWRITE_B6
    dd _HWRITE_B7
_HiROM_Read_No_SRAM_ROM_30:
    dd _READ_00_3F_No_SRAM
_HiROM_Read_No_SRAM_ROM_B0:
    dd _READ_80_BF_No_SRAM
_HiROM_Read_No_SRAM_Blank_30:
    dd _BlankH_00_3F_No_SRAM
_HiROM_Read_No_SRAM_Blank_B0:
    dd _BlankH_80_BF_No_SRAM
_HiROM_Write_No_SRAM_30:
    dd _HWRITE_00_3F_No_SRAM
_HiROM_Write_No_SRAM_B0:
    dd _HWRITE_80_BF_No_SRAM

section .text
; ebx - contains the actual address, al is where the info should be stored, edi is free
; NB ebx is not corrupted! edi is corrupted!
; NB eax is not corrupted barring returnvalue in al... e.g. ah should not be used etc!

%macro _HREAD_30_3F 1 ;bank
global _HREAD_%1
ALIGNC
_HREAD_%1:
    test bh,bh
    jns short %%not_rom
    ROM00_READ_32k
ALIGNC
%%not_rom:
    cmp bh,0x20
    jnb short %%not_lowram
    LOW_RAM_READ
ALIGNC
%%not_lowram:
    cmp bh,0x60
    jnb short %%not_ppu
    PPU_READ
ALIGNC
%%not_ppu:
    SRAM30_READ %1
%endmacro

global _HREAD_40_5F
ALIGNC
_HREAD_40_5F:
    HIROM40_LIN_READ

%macro _HREAD_B0_BF 1 ;bank
global _HREAD_%1
ALIGNC
_HREAD_%1:
    test bh,bh
    jns short %%not_rom
    ROM80_READ_32k
ALIGNC
%%not_rom:
    cmp bh,0x20
    jnb short %%not_lowram
    LOW_RAM_READ
ALIGNC
%%not_lowram:
    cmp bh,0x60
    jnb short %%not_ppu
    PPU_READ
ALIGNC
%%not_ppu:
    SRAMB0_READ %1
%endmacro

global _HREAD_C0_FF
ALIGNC
_HREAD_C0_FF:
    HIROMC0_LIN_READ

; Banks 30-3F are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF SRAM 8000-FFFF ROM
    _HREAD_30_3F 30
    _HREAD_30_3F 31
    _HREAD_30_3F 32
    _HREAD_30_3F 33
    _HREAD_30_3F 34
    _HREAD_30_3F 35
    _HREAD_30_3F 36
    _HREAD_30_3F 37
; Banks B0-BF are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF SRAM 8000-FFFF ROM
    _HREAD_B0_BF B0
    _HREAD_B0_BF B1
    _HREAD_B0_BF B2
    _HREAD_B0_BF B3
    _HREAD_B0_BF B4
    _HREAD_B0_BF B5
    _HREAD_B0_BF B6
    _HREAD_B0_BF B7

%macro _HWRITE_30_3F 1  ;bank
global _HWRITE_%1
ALIGNC
_HWRITE_%1:
    cmp bh,0x20
    jnb short %%not_lowram
    LOW_RAM_WRITE
ALIGNC
%%not_lowram:
    cmp bh,0x60
    jnb short %%not_ppu
    PPU_WRITE
ALIGNC
%%not_ppu:
    test bh,bh
    js  short %%not_sram
    SRAM30_WRITE %1
ALIGNC
%%not_sram:
    ROM_WRITE
%endmacro

%macro _HWRITE_B0_BF 1  ;bank
global _HWRITE_%1
ALIGNC
_HWRITE_%1:
    cmp bh,0x20
    jnb short %%not_lowram
    LOW_RAM_WRITE
ALIGNC
%%not_lowram:
    cmp bh,0x60
    jnb short %%not_ppu
    PPU_WRITE
ALIGNC
%%not_ppu:
    test bh,bh
    js  short %%not_sram
    SRAMB0_WRITE %1
ALIGNC
%%not_sram:
    ROM_WRITE
%endmacro

; Banks 30-3F are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF SRAM 8000-FFFF ROM
    _HWRITE_30_3F 30
    _HWRITE_30_3F 31
    _HWRITE_30_3F 32
    _HWRITE_30_3F 33
    _HWRITE_30_3F 34
    _HWRITE_30_3F 35
    _HWRITE_30_3F 36
    _HWRITE_30_3F 37
; Banks B0-BF are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF SRAM 8000-FFFF ROM
    _HWRITE_B0_BF B0
    _HWRITE_B0_BF B1
    _HWRITE_B0_BF B2
    _HWRITE_B0_BF B3
    _HWRITE_B0_BF B4
    _HWRITE_B0_BF B5
    _HWRITE_B0_BF B6
    _HWRITE_B0_BF B7

section .text; ALIGNC
section .data; ALIGND
section .bss; ALIGND
