;
;
; SNEeSe SNES memory mapper v1.6 - NASM rewrite
;
; Notes: LoROM has max ROM address space of 4MBytes.
;
; LoROM version
;  00-3F:0000-1FFF = Low RAM 8k (7E:0000-1FFF mirrored)
;  00-3F:2000-5FFF = Hardware ports (PPU, FX, DSP)
;  00-3F:6000-7FFF = Unmapped
;  00-3F:8000-FFFF = ROM 32k (0-16Mbit)
;  40-5F:8000-FFFF = ROM 32k (16-24Mbit)
;  60-6F:0000-FFFF = Unmapped
;  70-73:0000-FFFF = SRAM (up to 64k) (continuously mirrored: see below)
;  74-7D:0000-FFFF = Unmapped
;  7E-7F:0000-FFFF = WRAM 128k
;  80-DF:0000-FFFF = 00-5F mirrored
;  E0-FF:8000-FFFF = ROM 32k (24-32Mbit)
;
;  SRAM has 256k (kbytes, not kbits) of address space. This is mapped
;   with 'continuous mirroring', as such:
;    2k (16kbit): 70:0000-07FF == 70:0800-0FFF == 73:F800-FFFF
;    8k (64kbit): 70:0000-1FFF == 70:2000-3FFF == 73:E000-FFFF
;   etc.
;  Since I know of no ROMs that require >64k of SRAM, only 64k is
;   supported (however, 256k mirroring is performed).
;  Note: I HAVE discovered one ROM requiring 128k of SRAM (Dezaemon),
;   will implement this later...
;

%include "misc.ni"
%include "memmap.ni"

EXPORT_C memlo_start

extern UNSUPPORTED_WRITE
extern _RAM_WRITE
extern _WRITE_00_3F,_WRITE_80_BF
extern _WRAM,_SRAM_Mask,_SRAM
extern _RomReadAdjust2,_RomHiReadAdjust2
extern Read_Map_20_5F

section .data
; _LoROM_Read is _LoROM_Blank - see MemBlank.asm

global _LoROM_Write
ALIGND
_LoROM_Write:

; Banks 00-3F are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF RESERVED 8000-FFFF ROM
    DUPLICATE dd,64,_WRITE_00_3F

; Banks 40-5F are 0000-7FFF RESERVED 8000-FFFF ROM
    DUPLICATE dd,32,UNSUPPORTED_WRITE

; Banks 60-6F are not mapped!
    DUPLICATE dd,16,UNSUPPORTED_WRITE

; Banks 70-73 is SRAM (mirrored)
    DUPLICATE dd,4,_SRAM_WRITE

; Banks 74-7D are not mapped!
    DUPLICATE dd,10,UNSUPPORTED_WRITE

; Bank  7E-7f is  0000-FFFF RAM Nb 7E is used to store that 8KB block thats repeated in mem
    DUPLICATE dd,2,_RAM_WRITE

; Banks 80-BF are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF RESERVED 8000-FFFF ROM
    DUPLICATE dd,64,_WRITE_80_BF

; Banks C0-DF are 0000-7FFF RESERVED 8000-FFFF ROM
    DUPLICATE dd,32,UNSUPPORTED_WRITE

; Rest are unmapped
    DUPLICATE dd,32,UNSUPPORTED_WRITE

section .text
; ebx - contains the actual address, al is where the info should be stored, edi is free
; NB ebx is not corrupted! edi is corrupted!
; NB eax is not corrupted barring returnvalue in al... e.g. ah should not be used etc!

global _SRAM_READ
ALIGNC
_SRAM_READ:
    mov edi,[_SRAM_Mask]
    and edi,ebx
    mov al,[_SRAM+edi]
    ret

global _SRAM_WRITE
ALIGNC
_SRAM_WRITE:
    mov edi,[_SRAM_Mask]
    and edi,ebx
    mov [_SRAM+edi],al
    ret

; ebx - contains the actual address, al is where the info should be stored, edi is free
; NB ebx is not corrupted! edi is corrupted!
; NB eax is not corrupted barring returnvalue in al... e.g. ah should not be used etc!

global _READ_00_3F_2
ALIGNC
_READ_00_3F_2:
    test bh,bh
    jns short .not_rom
    ROM00_READ_32k_2
ALIGNC
.not_rom:
    cmp bh,0x20
    jnb short .not_lowram
    LOW_RAM_READ
ALIGNC
.not_lowram:
    cmp bh,0x60
    jnb short .not_ppu
    PPU_READ
ALIGNC
.not_ppu:
    NON_MAPPED_READ

global _READ_40_5F_2    ; v0.17: corrected 16-24Mbit
ALIGNC
_READ_40_5F_2:
    test bh,bh
    jns short .not_rom
    ROM00_READ_32k_2
ALIGNC
.not_rom:
    NON_MAPPED_READ

global _READ_80_BF_2
ALIGNC
_READ_80_BF_2:
    test bh,bh
    jns short .not_rom
    ROM80_READ_32k_2
ALIGNC
.not_rom:
    cmp bh,0x20
    jnb short .not_lowram
    LOW_RAM_READ
ALIGNC
.not_lowram:
    cmp bh,0x60
    jnb short .not_ppu
    PPU_READ
ALIGNC
.not_ppu:
    NON_MAPPED_READ

global _READ_C0_FF_2    ; v0.17: corrected 16-32Mbit
ALIGNC
_READ_C0_FF_2:
    test bh,bh
    jns short .not_rom
    ROM80_READ_32k_2
ALIGNC
.not_rom:
    NON_MAPPED_READ

section .text; ALIGNC
section .data; ALIGND
section .bss; ALIGND
