%ifndef SNEeSe_memmap_i
%define SNEeSe_memmap_i

%include "misc.ni"

; Read Low RAM - 0000-1FFF in 00-3F/80-BF
%macro LOW_RAM_READ 0
    mov edi,0x1FFF
    and edi,ebx
    mov al,[_WRAM+edi]
    ret
%endmacro

; Write Low RAM - 0000-1FFF in 00-3F/80-BF
%macro LOW_RAM_WRITE 0
    mov edi,0x1FFF
    and edi,ebx
    mov [_WRAM+edi],al
    ret
%endmacro

; Read hardware - 2000-5FFF in 00-3F/80-BF
%macro PPU_READ 0
    mov edi,0xFFFF
    and edi,ebx
    jmp [(Read_Map_20_5F-0x2000*4)+edi*4]
%endmacro

; Write hardware - 2000-5FFF in 00-3F/80-BF
%macro PPU_WRITE 0
    mov edi,0xFFFF
    and edi,ebx
    jmp [(Write_Map_20_5F-0x2000*4)+edi*4]
%endmacro

; Read hardware - 2100-21FF in 00-3F/80-BF
%macro PPU_21_READ 0
    mov edi,0xFF
    and edi,ebx
    jmp [Read_Map_21+edi*4]
%endmacro

; Write hardware - 2100-21FF in 00-3F/80-BF
%macro PPU_21_WRITE 0
    mov edi,0xFF
    and edi,ebx
    jmp [Write_Map_21+edi*4]
%endmacro

; Read hardware - 4000-40FF in 00-3F/80-BF
%macro PPU_40_READ 0
    mov edi,0xFF
    and edi,ebx
    jmp [Read_Map_40+edi*4]
%endmacro

; Write hardware - 4000-40FF in 00-3F/80-BF
%macro PPU_40_WRITE 0
    mov edi,0xFF
    and edi,ebx
    jmp [Write_Map_40+edi*4]
%endmacro

; Read hardware - 4200-42FF in 00-3F/80-BF
%macro PPU_42_READ 0
    mov edi,0xFF
    and edi,ebx
    jmp [Read_Map_42+edi*4]
%endmacro

; Write hardware - 4200-42FF in 00-3F/80-BF
%macro PPU_42_WRITE 0
    mov edi,0xFF
    and edi,ebx
    jmp [Write_Map_42+edi*4]
%endmacro

; Read hardware - 4300-43FF in 00-3F/80-BF
%macro PPU_43_READ 0
    mov edi,0xFF
    and edi,ebx
    jmp [Read_Map_43+edi*4]
%endmacro

; Write hardware - 4300-43FF in 00-3F/80-BF
%macro PPU_43_WRITE 0
    mov edi,0xFF
    and edi,ebx
    jmp [Write_Map_43+edi*4]
%endmacro

; Read LoROM ROM 00-5F (first half of banks)
; Read HiROM ROM 00-3F
%macro ROM00_READ_32k 0
    mov edi,[_RomReadAdjust]
    mov al,[ebx+edi]
    ret
%endmacro

; Read LoROM ROM 00-5F (second half of banks)
%macro ROM00_READ_32k_2 0
    mov edi,[_RomReadAdjust2]
    mov al,[ebx+edi]
    ret
%endmacro

; Read LoROM ROM 80-FF (first half of banks)
; Read HiROM ROM 80-BF
%macro ROM80_READ_32k 0
    mov edi,[_RomHiReadAdjust]
    mov al,[ebx+edi]
    ret
%endmacro

; Read LoROM ROM 80-FF (second half of banks)
%macro ROM80_READ_32k_2 0
    mov edi,[_RomHiReadAdjust2]
    mov al,[ebx+edi]
    ret
%endmacro

; Read HiROM ROM 40-5F
%macro HIROM40_LIN_READ 0
    mov edi,[_HRomLinReadAdjust]
    mov al,[ebx+edi]
    ret
%endmacro

; Read HiROM ROM C0-FF
%macro HIROMC0_LIN_READ 0
    mov edi,[_HRomHiLinReadAdjust]
    mov al,[ebx+edi]
    ret
%endmacro

; Read from ROM address space not containing ROM
%macro NO_ROM_READ 0
    mov al,0xFF
    ret
%endmacro

; Write to ROM address space
%macro ROM_WRITE 0
    ret
%endmacro

; Read HiROM SRAM 30-3F
%macro SRAM30_READ 1    ;num
    mov edi,[_SRAM_Mask]
    and edi,ebx
    mov al,[(_SRAM+((0x%1-0x30)*0x2000))+edi]
    ret
%endmacro

; Write HiROM SRAM 30-3F
%macro SRAM30_WRITE 1   ;num
    mov edi,[_SRAM_Mask]
    and edi,ebx
    mov [(_SRAM+((0x%1-0x30)*0x2000))+edi],al
    ret
%endmacro

; Read HiROM SRAM B0-BF
%macro SRAMB0_READ 1    ;num
    mov edi,[_SRAM_Mask]
    and edi,ebx
    mov al,[(_SRAM+((0x%1-0xB0)*0x2000))+edi]
    ret
%endmacro

; Write HiROM SRAM B0-BF
%macro SRAMB0_WRITE 1   ;num
    mov edi,[_SRAM_Mask]
    and edi,ebx
    mov [(_SRAM+((0x%1-0xB0)*0x2000))+edi],al
    ret
%endmacro

; Non-mapped/unsupported read
%macro NON_MAPPED_READ 0
    mov al,0xFF
    ret
%endmacro

; Non-mapped/unsupported write
%macro NON_MAPPED_WRITE 0
    ret
%endmacro

%endif ;!SNEeSe_memmap_i
