/*

 Mode 7 matrix rendering / hardware port emulation.

*/

#include "misc.i"
#include "spriteL.i"
#include "screen.i"

EXPORT_C mode7_start

.data
ALIGND
.globl M7_Table
M7_Table:
.long M7_REPEAT,M7_REPEAT_X,M7_REPEAT_Y,M7_REPEAT_XY
.long M7_CLIP,M7_CLIP_X,M7_CLIP_Y,M7_CLIP_XY
.long M7_CLIP,M7_CLIP_X,M7_CLIP_Y,M7_CLIP_XY
.long M7_CHAR0,M7_CHAR0_X,M7_CHAR0_Y,M7_CHAR0_XY
.long M7P_REPEAT,M7P_REPEAT_X,M7P_REPEAT_Y,M7P_REPEAT_XY
.long M7P_CLIP,M7P_CLIP_X,M7P_CLIP_Y,M7P_CLIP_XY
.long M7P_CLIP,M7P_CLIP_X,M7P_CLIP_Y,M7P_CLIP_XY
.long M7P_CHAR0,M7P_CHAR0_X,M7P_CHAR0_Y,M7P_CHAR0_XY

.bss
ALIGND
.globl Mode7_A_P,Mode7_B,Mode7_C_P,Mode7_D
Mode7_A_P:  skipl   // Partially precalculated
Mode7_B:    skipl
Mode7_C_P:  skipl   // Partially precalculated
Mode7_D:    skipl

.globl _M7A,_M7B,_M7C,_M7D,_M7X_13,_M7Y_13,_M7H,_M7V,_M7X,_M7Y
_M7A:   skipl
_M7B:   skipl
_M7C:   skipl
_M7D:   skipl
_M7X_13:skipl
_M7Y_13:skipl
_M7H:   skipl
_M7V:   skipl
_M7X:   skipl
_M7Y:   skipl

MPYL:   skipb   // Mode 7 multiplication result: low byte
MPYM:   skipb   // Mode 7 multiplication result: middle byte
MPYH:   skipb   // Mode 7 multiplication result: high byte
        skipb

.globl M7_Handler
M7_Handler: skipl
.globl _M7SEL
_M7SEL:     skipb   // ab0000yx  ab=mode 7 repetition info,y=flip vertical,x=flip horizontal
.globl Redo_M7
Redo_M7:    skipb
Redo_16x8:  skipb

.text
ALIGNC
.globl RENDER_MODE_7
RENDER_MODE_7:
    CALCULATE_MOSAIC_YLINE 0x01

    movl ScreenDestPtr,%edi

    movb Redo_M7,%dl
    testb $0x0F,%dl // Need to do any recalculating?
    jz  0f
    testb $0x0A,%dl // Recalculate B or D?
    jz  3f
    movl _M7V,%eax
    subl _M7Y_13,%eax
    testb $0x02,%dl // Recalculate B?
    jz  1f
    pushl %eax
    imull _M7B,%eax
    movl %eax,Mode7_B
    popl %eax
    testb $0x08,%dl // Recalculate D?
    jz  2f
1:  imull _M7D,%eax
    movl %eax,Mode7_D
2:  testb $0x05,%dl // Recalculate A or C?
    jz  0f
3:  movl _M7H,%eax
    subl _M7X_13,%eax
    testb $0x01,%dl // Recalculate A?
    jz  1f
    pushl %eax
    imull _M7A,%eax
    movl _M7X_13,%ebx
    shll $8,%ebx
    addl %ebx,%eax
    movl %eax,Mode7_A_P
    popl %eax
    testb $0x04,%dl // Recalculate C?
    jz  0f
1:  imull _M7C,%eax
    movl _M7Y_13,%ebx
    shll $8,%ebx
    addl %ebx,%eax
    movl %eax,Mode7_C_P
0:  movb $0,Redo_M7 // Done with recalculating

    movb MosaicBG1,%al
    testb %al,%al
    movl Current_Line_Render,%ebx
    jz  0f
    movl Current_Line_Mosaic,%ebx
0:
    movl $_VRAM_DI,%esi
    movl $256,%ebp      // Horizontal Count
    movl _M7B,%eax
    movl _M7A,%ecx
    movl _M7C,%edx
    jmpl *M7_Handler    // New for 0.25

.macro M7__REPEAT xflip,priority,label,ylabel
ALIGNC
\ylabel:
    xorb $0xFF,%bl
\label:
.if \xflip-0
    addl $255,%edi
.endif
    imull %ebx,%eax
    imull _M7D,%ebx
    addl Mode7_A_P,%eax
    addl Mode7_C_P,%ebx
    addl Mode7_B,%eax
.if \priority-0
    movb Tile_priority_opcode,%cl
.endif
    addl Mode7_D,%ebx
.if \priority-0
    movb %cl,priority_\label
.endif

ALIGNC
1:  cmpb $0,(%edi)
    jnz 9f

    pushl %eax
    pushl %ebx

    shrl $8,%eax
    shrl $8,%ebx
    andl $0x3FF,%eax
    andl $0x3FF,%ebx

// Convert Screen X,Y location to SNES Pic location

                // Assumes %ax is X coord 0-1023, %bx is Y 0-1023
    movl %eax,%ecx
    movl %ebx,%edx
    shrl $3,%eax    // Get Tile Position (in 128 by 128 map)
    andl $7,%ecx    // Get pixel shift within tile
    shll $5,%ebx    // Tile Position*256 cos thats width of map
    andl $7,%edx
    addb %al,%al
    movb %al,%bl    // Map offset now in %ebx
    shrl $1,%ebx
    xorl %eax,%eax
    shll $3,%edx
    movb (%esi,%ebx),%al    // Got Tile Number
    
    shll $6,%eax            // Get offset to tile data
    addl %ecx,%eax          // Add X offset
    addl %edx,%eax          // Add Y offset

    movb 32768(%esi,%eax),%al
.if \priority-0
    testb %al,%al
priority_\label:
    js  3f
    andb $0x7F,%al
.endif
    movb %al,(%edi)
3:  popl %ebx
    popl %eax
    movl _M7A,%ecx
    movl _M7C,%edx

9:  addl %ecx,%eax
    addl %edx,%ebx

.if \xflip-0
    decl %edi
.else
    incl %edi
.endif

    decl %ebp
    jnz 1b

    ret
.endm

// New for v0.16, tile 0 repeat support
.macro M7__CHAR0 xflip,priority,label,ylabel
ALIGNC
\ylabel:
    xorb $0xFF,%bl
\label:
.if \xflip-0
    addl $255,%edi
.endif
    imull %ebx,%eax
    imull _M7D,%ebx
    addl Mode7_A_P,%eax
    addl Mode7_C_P,%ebx
    addl Mode7_B,%eax
.if \priority-0
    movb Tile_priority_opcode,%cl
.endif
    addl Mode7_D,%ebx
.if \priority-0
    movb %cl,priority_\label
.endif

ALIGNC
1:  cmpb $0,(%edi)
    jnz 9f

    pushl %eax
    pushl %ebx

    shrl $8,%eax
    shrl $8,%ebx

// Convert Screen X,Y location to SNES Pic location

                // Assumes %ax is X coord 0-1023, %bx is Y 0-1023
    movl %eax,%ecx
    movl %ebx,%edx
    andl $7,%ecx    // Get pixel shift within tile
    andl $7,%edx

    cmpl $1023,%eax // If outside screen range we use tile 0
    ja 2f
    cmpl $1023,%ebx
    jbe 3f
2:  xorl %eax,%eax
    jmp 4f

ALIGNC
3:  shrl $3,%eax    // Get Tile Position (in 128 by 128 map)
    shll $5,%ebx    // Tile Position*256 cos thats width of map
    addb %al,%al
    movb %al,%bl    // Map offset now in %ebx
    shrl $1,%ebx
    xorl %eax,%eax
    movb (%esi,%ebx),%al    // Got Tile Number
    
    shll $6,%eax            // Get offset to tile data
4:  addl %ecx,%eax          // Add X offset
    shll $3,%edx
    addl %edx,%eax          // Add Y offset

    movb 32768(%esi,%eax),%al
.if \priority-0
    testb %al,%al
priority_\label:
    js  3f
    andb $0x7F,%al
.endif
    movb %al,(%edi)
3:  popl %ebx
    popl %eax
    movl _M7A,%ecx
    movl _M7C,%edx

9:  addl %ecx,%eax
    addl %edx,%ebx

.if \xflip-0
    decl %edi
.else
    incl %edi
.endif

    decl %ebp
    jnz 1b

    ret
.endm

.macro M7__CLIP xflip,priority,label,ylabel
ALIGNC
\ylabel:
    xorb $0xFF,%bl
\label:
.if \xflip-0
    addl $255,%edi
.endif
    imull %ebx,%eax
    imull _M7D,%ebx
    addl Mode7_A_P,%eax
    addl Mode7_C_P,%ebx
    addl Mode7_B,%eax
.if \priority-0
    movb Tile_priority_opcode,%cl
.endif
    addl Mode7_D,%ebx
.if \priority-0
    movb %cl,priority_\label
.endif

ALIGNC
1:  cmpb $0,(%edi)
    jnz 9f

    testl $~0x3FFFF,%eax  // If outside screen range we simply skip the pixel
    jnz 9f

    testl $~0x3FFFF,%ebx
    jnz 9f

// Convert Screen X,Y location to SNES Pic location

                // Assumes %ax is X coord 0-1023, %bx is Y 0-1023
    pushl %eax
    pushl %ebx
    shrl $8,%eax
    shrl $8,%ebx
    movl %eax,%ecx
    movl %ebx,%edx
    shrl $3,%eax    // Get Tile Position (in 128 by 128 map)
    andl $7,%ecx    // Get pixel shift within tile
    shll $5,%ebx    // Tile Position*256 cos thats width of map
    andl $7,%edx
    addb %al,%al
    movb %al,%bl    // Map offset now in %ebx
    shrl $1,%ebx
    xorl %eax,%eax
    shll $3,%edx
    movb (%esi,%ebx),%al    // Got Tile Number
    
    shll $6,%eax            // Get offset to tile data
    addl %ecx,%eax          // Add X offset
    addl %edx,%eax          // Add Y offset

    movb 32768(%esi,%eax),%al
.if \priority-0
    testb %al,%al
priority_\label:
    js  3f
    andb $0x7F,%al
.endif
    movb %al,(%edi)
3:  popl %ebx
    popl %eax
    movl _M7A,%ecx
    movl _M7C,%edx

9:  addl %ecx,%eax
    addl %edx,%ebx

.if \xflip-0
    decl %edi
.else
    incl %edi
.endif

    decl %ebp
    jnz 1b

    ret
.endm

M7__REPEAT 0, 0, M7_REPEAT, M7_REPEAT_Y
M7__REPEAT 1, 0, M7_REPEAT_X, M7_REPEAT_XY
M7__CHAR0  0, 0, M7_CHAR0, M7_CHAR0_Y
M7__CHAR0  1, 0, M7_CHAR0_X, M7_CHAR0_XY
M7__CLIP   0, 0, M7_CLIP, M7_CLIP_Y
M7__CLIP   1, 0, M7_CLIP_X, M7_CLIP_XY

M7__REPEAT 0, 1, M7P_REPEAT, M7P_REPEAT_Y
M7__REPEAT 1, 1, M7P_REPEAT_X, M7P_REPEAT_XY
M7__CHAR0  0, 1, M7P_CHAR0, M7P_CHAR0_Y
M7__CHAR0  1, 1, M7P_CHAR0_X, M7P_CHAR0_XY
M7__CLIP   0, 1, M7P_CLIP, M7P_CLIP_Y
M7__CLIP   1, 1, M7P_CLIP_X, M7P_CLIP_XY

.globl SCREEN_MODE_7
SCREEN_MODE_7:  // There is 1 version of the render code!
  testb $0x10,%bl   // Are sprites enabled?
  jz 1f

  movb Redo_OAM,%al
  testb %al,%al
  jz 0f
  pushl %ebx
  pusha
  call Recache_OAM
  popa
  popl %ebx
0:andl $3,%ebx
  jz NO_PLANES
  andl $1,%ebx
  jnz M7_P1
  // we only reach here when bit 0 clear, bit 1 set
  testl $0x40,_SETINI
  jnz M7P_P1
0:ret

ALIGNC
1:andl $3,%ebx
  jz 0f
  andl $1,%ebx
  jnz RENDER_MODE_7
  // we only reach here when bit 0 clear, bit 1 set
  testl $0x40,_SETINI
  jnz RENDER_MODE_7
0:ret

ALIGNC
M7_P1:
  ALL_SPRITES_PLANE 3
  ALL_SPRITES_PLANE 2
  ALL_SPRITES_PLANE 1
  call RENDER_MODE_7
  ALL_SPRITES_PLANE 0
  ret

ALIGNC
M7P_P1:
  ALL_SPRITES_PLANE 3
  ALL_SPRITES_PLANE 2
  ALL_SPRITES_PLANE 1
  movb $0x79,Tile_priority_opcode //jns
  call RENDER_MODE_7
  ALL_SPRITES_PLANE 0
  movb $0x78,Tile_priority_opcode //js
  call RENDER_MODE_7
  ret

ALIGNC
.globl _Reset_Mode_7
_Reset_Mode_7:
  // Set eax to 0, as we're setting most everything to 0...
  xorl %eax,%eax

  movb %al,_M7SEL
  movb $0x0F,Redo_M7
  movl $M7_REPEAT,M7_Handler
  movb $0,Redo_16x8
  movl %eax,MPYL
  movl %eax,_M7A
  movl %eax,_M7B
  movl %eax,_M7C
  movl %eax,_M7D
  movl %eax,_M7X_13
  movl %eax,_M7Y_13
  movl %eax,_M7H
  movl %eax,_M7V
  movl %eax,_M7X
  movl %eax,_M7Y

  ret

.globl SNES_R2134,SNES_R2135,SNES_R2136,SNES_W211A,SNES_W211B
.globl SNES_W211C,SNES_W211D,SNES_W211E,SNES_W211F,SNES_W2120
Do_16x8_Multiply:
  pushl %ebx
  movl %edi,%ebx
  movsbl _M7B+1,%edi
  movb $0,Redo_16x8
  imull _M7A,%edi   // I think signed is used makes most sense!
  movl %edi,MPYL
  movb (%ebx),%al
  popl %ebx
  ret

ALIGNC
SNES_R2134: // MPYL
  movl $MPYL,%edi
  cmpb $0,Redo_16x8
  jnz Do_16x8_Multiply
  movb (%edi),%al
  ret

ALIGNC
SNES_R2135: // MPYM
  movl $MPYM,%edi
  cmpb $0,Redo_16x8
  jnz Do_16x8_Multiply
  movb (%edi),%al
  ret

ALIGNC
SNES_R2136: // MPYH
  movl $MPYH,%edi
  cmpb $0,Redo_16x8
  jnz Do_16x8_Multiply
  movb (%edi),%al
  ret

ALIGNC
SNES_W211A: // M7SEL    // New for 0.12
  cmpb _M7SEL,%al
  je 0f
  SyncRender    //*
  // Note: M7_Table,M7_Handler are declared in ScreenL.S
  pushl %ebx
  movb %al,%bl
  movb %al,_M7SEL
  shrl $4,%ebx
  andb $3,%al
  andl $0x0C,%ebx
  orb %al,%bl

  movb _SETINI,%al  // 0.42 - Mode 7 w/ priority
  andb $0x40,%al
  shrb $2,%al
  orb %al,%bl

  movl M7_Table(,%ebx,4),%ebx
  movl %ebx,M7_Handler
  popl %ebx
  movb _M7SEL,%al
0:ret

ALIGNC
SNES_W211B: // M7A
  SyncRender    //*M7
// Used for matrix render and 16-bit M7A * 8-bit = 24-bit multiply
  pushl %eax
  movb %al,%ah
  movb _M7A+1,%al
  cwde
  movl %eax,_M7A
  movb $0x01,%al    // Recalculate A
  orb  %al,Redo_M7
  movb $-1,Redo_16x8
  popl %eax
  ret

ALIGNC
SNES_W211C: // M7B
  SyncRender    //*M7
// Used for matrix render and 16-bit * 8-bit M7B high byte = 24-bit multiply
  pushl %eax
  movb %al,%ah
  movb _M7B+1,%al
  cwde
  movl %eax,_M7B
  movb $0x02,%al    // Recalculate B
  orb  %al,Redo_M7
  movb $-1,Redo_16x8
  popl %eax
  ret

ALIGNC
SNES_W211D: // M7C
  SyncRender    //*M7
  pushl %eax
  movb %al,%ah
  movb _M7C+1,%al
  cwde
  movl %eax,_M7C
  movb $0x04,%al    // Recalculate C
  orb  %al,Redo_M7
  popl %eax
  ret

ALIGNC
SNES_W211E: // M7D
  SyncRender    //*M7
  pushl %eax
  movb %al,%ah
  movb _M7D+1,%al
  cwde
  movl %eax,_M7D
  movb $0x08,%al    // Recalculate D
  orb  %al,Redo_M7
  popl %eax
  ret

ALIGNC
SNES_W211F: // M7X
  SyncRender    //*M7
  pushl %eax
  movb %al,%ah
  movb _M7X+1,%al
  movw %ax,_M7X

  shll $0x13,%eax
  sarl $0x13,%eax
  movl %eax,_M7X_13
  movb $0x05,%al    // Recalculate A,C
  orb  %al,Redo_M7
  popl %eax
  ret

ALIGNC
SNES_W2120: // M7Y
  SyncRender    //*M7
  pushl %eax
  movb %al,%ah
  movb _M7Y+1,%al
  movw %ax,_M7Y

  shll $0x13,%eax
  sarl $0x13,%eax
  movl %eax,_M7Y_13
  movb $0x0E,%al    // Recalculate B,C,D
  orb  %al,Redo_M7
  popl %eax
  ret

.text; ALIGNC
.data; ALIGND
.bss; ALIGND
