#if !(defined(__cplusplus)||defined(c_plusplus))
#error romload.h requires C++
#else
#ifndef SNEeSe_romload_h
#define SNEeSe_romload_h

/*
 ROM Copier Header
*/
struct ROM_Header {
 unsigned short NumPages;
 unsigned char ImageInformation;
 char Reserved_1[5];
 unsigned char SWC_Ident1;
 unsigned char SWC_Ident2;
 unsigned char SWC_Ident3;
 char RestOfHeader[501];
};

typedef enum {
 ROMFileType_normal, ROMFileType_split,
 ROMFileType_gamedoctor, ROMFileType_compressed
} ROMFileType;

const int
 Undetected = -1, Off = 0, On = 1,
 LoROM = 0, HiROM = 1,
 Non_Interleaved = 0,Interleaved = 2,
 LoROM_Interleaved = LoROM | Interleaved,
 HiROM_Interleaved = HiROM | Interleaved,
 NTSC_video = 0, PAL_video = 1;

#define ROM_Header_Size (sizeof(ROM_Header))

/*
 ROM Info Block
*/
struct SNESRomInfoStruct {
 char ROM_Title[21];           //FFC0
 unsigned char ROM_makeup;     //FFD6
 unsigned char ROM_type;       //FFD7
 unsigned char ROM_size;       //FFD8
 unsigned char SRAM_size;      //FFD9
 unsigned char Country_code;   //FFDA
 unsigned char License;        //FFDB
 unsigned char ROM_version;    //FFDC
 unsigned short int Complement;//FFDD
 unsigned short int Checksum;  //FFDF
};

#ifndef TRUE
#define TRUE (-1)
#endif
#ifndef FALSE
#define FALSE (0)
#endif

extern char *rom_romfile;
extern char *rom_romhilo;
extern char *rom_romname;
extern char *rom_romtype;
extern char *rom_romsize;
extern char *rom_sram;
extern char *rom_country;

void Timer_Handler_Throttle(...);
void Reset_Memory(void);
extern unsigned char *RomAddress;       // Address of SNES ROM

// Used to determine size of file for saving/loading, and to restrict writes
//  to non-existant SRAM
extern unsigned long SaveRamLength;

extern int Allocate_Rom(void);
extern unsigned NumBanks;
extern int ROM_format;
extern bool ROM_Loaded;

extern SNESRomInfoStruct RomInfoLo,RomInfoHi;

int SaveSRAM(char *SRAM_filename);
int LoadSRAM(char *SRAM_filename);
extern char *ROM_filename;
extern char fn_drive[MAXDRIVE], fn_dir[MAXDIR], fn_file[MAXFILE], fn_ext[MAXEXT];
extern char SRAM_filename[MAXPATH];
extern char save_dir[MAXPATH];
extern char save_extension[4];
void CreateSRAMFilename(char *ROM_filename);

int Setup_Memory(void);
int Allocate_ROM(void);

extern int ROM_has_header;
/* Off = LoROM, On = HiROM, Undetected = autodetect */
extern int ROM_memory_map;
extern int ROM_interleaved;
extern int ROM_video_standard;

int LoadROM(char *FileName);
void DisplayRomStats(SNESRomInfoStruct *RomInfo);

#ifdef DEBUG
#ifdef TRACKERS
extern unsigned char *HDMAAddress;
extern unsigned long LastHDMA;
extern unsigned char *InsAddress;
extern unsigned long LastIns;
#endif
#endif

extern char *TypeTable[];
extern char *CountryTable[];

extern unsigned long Frametime;

extern void Wangle(void);
#endif /* SNEeSe_romload_h */
#endif /* defined(__cplusplus)||defined(c_plusplus) */
