/* SPC instruction macros
 *  Original code by Savoury SnaX
 *  Rewritten/maintained by TRAC
 */

/* TCALL: Call through table in pageable 64-bytes of RAM */
.macro TCALL vecnum
ALIGNC
SPC_TCALL_\vecnum:
  GET_PC %eax
  incl %eax
  PUSH_W
  xorl %eax,%eax
  movw _SPCRAM+0xFFC0+((15 - (\vecnum)) << 1),%ax
  cmpl $0xFFC0,%eax
  movl %eax,_NativePC
  movl SPC_FFC0_Address,%eax
  jnb 0f
  movl $_SPCRAM,%eax
0:addl %eax,_NativePC
  movl %eax,SPC_Code_Base
  jmp SPC_BACK
.endm

.macro SPC_BFC label,flag
ALIGNC
SPC_\label:
    addl $2,_NativePC
    JUMP_FLAG_SPC \flag,SPC_BACK        // flag set
    movsbl -1(_NativePC),%eax           // sign extend for addition
    addl $2,_Cycles                     // Branch taken
//  addl $2,_TotalCycles    // (0.32) Butcha - fix 'lost' SPC timer ticks!
    addl %eax,_NativePC
    jmp SPC_BACK
.endm

.macro SPC_BFS label,flag
ALIGNC
SPC_\label:
    addl $2,_NativePC
    JUMP_NOT_FLAG_SPC \flag,SPC_BACK    // flag set
    movsbl -1(_NativePC),%eax           // sign extend for addition
    addl $2,_Cycles                     // Branch taken
//  addl $2,_TotalCycles    // (0.32) Butcha - fix 'lost' SPC timer ticks!
    addl %eax,_NativePC
    jmp SPC_BACK
.endm

.macro SPC_OP_A_IM op,flags
    movb 1(_NativePC),%al
    addl $2,_NativePC
    SPC_BASE_\op %al,R_A
    STORE_FLAGS_SPC (\flags)
    jmp SPC_BACK
.endm

.macro SPC_OP_X_IM op,flags
    movb 1(_NativePC),%al
    addl $2,_NativePC
    SPC_BASE_\op %al,R_X
    STORE_FLAGS_SPC (\flags)
    jmp SPC_BACK
.endm

.macro SPC_OP_Y_IM op,flags
    movb 1(_NativePC),%al
    addl $2,_NativePC
    SPC_BASE_\op %al,R_Y
    STORE_FLAGS_SPC (\flags)
    jmp SPC_BACK
.endm

.macro SPC_OP_A op,flags,addr
    ADDR_\addr
    GET_BYTE_SPC
    SPC_BASE_\op %al,R_A
    STORE_FLAGS_SPC (\flags)
    jmp SPC_BACK
.endm

.macro SPC_OP_X op,flags,addr
    ADDR_\addr
    GET_BYTE_SPC
    SPC_BASE_\op %al,R_X
    STORE_FLAGS_SPC (\flags)
    jmp SPC_BACK
.endm

.macro SPC_OP_Y op,flags,addr
    ADDR_\addr
    GET_BYTE_SPC
    SPC_BASE_\op %al,R_Y
    STORE_FLAGS_SPC (\flags)
    jmp SPC_BACK
.endm

.macro SPC_OP_mem op,flags,addr
    ADDR_\addr 1,%cl    // %cl is source byte, %bx dest address
    GET_BYTE_SPC        // %al contains byte at dest address
    SPC_BASE_\op %cl,%al
    STORE_FLAGS_SPC (\flags)
    SET_BYTE_SPC
    jmp SPC_BACK
.endm

.macro SPC_BASE_MOV src,dest
    testb \src,\src
    movb \src,\dest
.endm

.macro SPC_BASE_OR src,dest
    orb \src,\dest
.endm

.macro SPC_OR_A addr
ALIGNC
SPC_OR_A_\addr:
    SPC_OP_A OR,SPC_FLAG_NZ,\addr
.endm

.macro SPC_OR_mem addr
ALIGNC
SPC_OR_\addr:
    SPC_OP_mem OR,SPC_FLAG_NZ,\addr
.endm

.macro SPC_BASE_AND src,dest
    andb \src,\dest
.endm

.macro SPC_AND_A addr
ALIGNC
SPC_AND_A_\addr:
    SPC_OP_A AND,SPC_FLAG_NZ,\addr
.endm

.macro SPC_AND_mem addr
ALIGNC
SPC_AND_\addr:
    SPC_OP_mem AND,SPC_FLAG_NZ,\addr
.endm

.macro SPC_BASE_EOR src,dest
    xorb \src,\dest
.endm

.macro SPC_EOR_A addr
ALIGNC
SPC_EOR_A_\addr:
    SPC_OP_A EOR,SPC_FLAG_NZ,\addr
.endm

.macro SPC_EOR_mem addr
ALIGNC
SPC_EOR_\addr:
    SPC_OP_mem EOR,SPC_FLAG_NZ,\addr
.endm

.macro SPC_BASE_CMP src,dest
    cmpb \src,\dest
    cmc
.endm

.macro SPC_CMP reg,addr
    ADDR_\addr
    GET_BYTE_SPC
    cmpb %al,\reg
    cmc
    STORE_FLAGS_SPC (SPC_FLAG_NZC)
    jmp SPC_BACK
.endm

.macro SPC_CMP_A addr
ALIGNC
SPC_CMP_A_\addr:
    SPC_CMP R_A,\addr
.endm

.macro SPC_CMP_X addr
ALIGNC
SPC_CMP_X_\addr:
    SPC_CMP R_X,\addr
.endm

.macro SPC_CMP_Y addr
ALIGNC
SPC_CMP_Y_\addr:
    SPC_CMP R_Y,\addr
.endm

.macro SPC_CMP_mem addr
ALIGNC
SPC_CMP_\addr:
    ADDR_\addr 1,%cl    // %cl is source byte, %bx dest address
    GET_BYTE_SPC        // %al contains byte at dest address
    cmpb %cl,%al
    cmc
    STORE_FLAGS_SPC (SPC_FLAG_NZC)
    jmp SPC_BACK
.endm

.macro SPC_ADC_A addr
ALIGNC
SPC_ADC_A_\addr:
    ADDR_\addr
    CLR_FLAG_SPC(SPC_FLAG_V)    // Done cos STORE_FLAGS cant alter V!
    GET_BYTE_SPC
    MAKE_CARRY
    adcb %al,R_A
    lahf
    jno 1f
    SET_FLAG_SPC(SPC_FLAG_V)
1:  STORE_FLAGS_SPC_FROM SPC_FLAG_NHZC,%ah
    jmp SPC_BACK
.endm

.macro SPC_ADC_mem addr
ALIGNC
SPC_ADC_\addr:
    ADDR_\addr 1,%cl
    CLR_FLAG_SPC(SPC_FLAG_V)    // Done cos STORE_FLAGS cant alter V!
    GET_BYTE_SPC
    MAKE_CARRY
    adcb %cl,%al
    lahf
    jno 1f
    SET_FLAG_SPC(SPC_FLAG_V)
1:  STORE_FLAGS_SPC_FROM SPC_FLAG_NHZC,%ah
    SET_BYTE_SPC
    jmp SPC_BACK
.endm

.macro SPC_SBC_A addr
ALIGNC
SPC_SBC_A_\addr:
    ADDR_\addr
    CLR_FLAG_SPC(SPC_FLAG_V)    // Done cos STORE_FLAGS cant alter V!
    GET_BYTE_SPC
    MAKE_NOT_CARRY
    sbbb %al,R_A
    cmc
    lahf
    jno 1f
    SET_FLAG_SPC(SPC_FLAG_V)
1:  STORE_FLAGS_SPC_FROM SPC_FLAG_NHZC,%ah
    jmp SPC_BACK
.endm

.macro SPC_SBC_mem addr
ALIGNC
SPC_SBC_\addr:
    ADDR_\addr 1,%cl
    CLR_FLAG_SPC(SPC_FLAG_V)    // Done cos STORE_FLAGS cant alter V!
    GET_BYTE_SPC
    MAKE_NOT_CARRY
    sbbb %cl,%al    // (0.30) Butcha - switched these, wrong order
    cmc
    lahf
    jno 1f
    SET_FLAG_SPC(SPC_FLAG_V)
1:  STORE_FLAGS_SPC_FROM SPC_FLAG_NHZC,%ah
    SET_BYTE_SPC
    jmp SPC_BACK
.endm

.macro SPC_DEC_reg reg
ALIGNC
SPC_DEC_\reg:
    incl _NativePC
    decb __SPC_\reg
    STORE_FLAGS_SPC (SPC_FLAG_NZ)
    jmp SPC_BACK
.endm

.macro SPC_DEC addr
ALIGNC
SPC_DEC_\addr:
    ADDR_\addr
    GET_BYTE_SPC
    decb %al
    STORE_FLAGS_SPC (SPC_FLAG_NZ)
    SET_BYTE_SPC
    jmp SPC_BACK
.endm

.macro SPC_INC_reg reg
ALIGNC
SPC_INC_\reg:
    incl _NativePC
    incb __SPC_\reg
    STORE_FLAGS_SPC (SPC_FLAG_NZ)
    jmp SPC_BACK
.endm

.macro SPC_INC addr
ALIGNC
SPC_INC_\addr:
    ADDR_\addr
    GET_BYTE_SPC
    incb %al
    STORE_FLAGS_SPC (SPC_FLAG_NZ)
    SET_BYTE_SPC
    jmp SPC_BACK
.endm

.macro SPC_ASL addr
ALIGNC
SPC_ASL_\addr:
    ADDR_\addr
    GET_BYTE_SPC
    addb %al,%al
    STORE_FLAGS_SPC (SPC_FLAG_NZC)
    SET_BYTE_SPC
    jmp SPC_BACK
.endm

.macro SPC_LSR addr
ALIGNC
SPC_LSR_\addr:
    ADDR_\addr
    GET_BYTE_SPC
    shrb $1,%al
    STORE_FLAGS_SPC (SPC_FLAG_NZC)
    SET_BYTE_SPC
    jmp SPC_BACK
.endm

.macro SPC_ROL addr
ALIGNC
SPC_ROL_\addr:
    ADDR_\addr
    GET_BYTE_SPC
    MAKE_CARRY
    adcb %al,%al
    STORE_FLAGS_SPC (SPC_FLAG_NZC)
    SET_BYTE_SPC
    jmp SPC_BACK
.endm

.macro SPC_ROR addr
ALIGNC
SPC_ROR_\addr:
    ADDR_\addr
    GET_BYTE_SPC
    shrb $1,_PSW_B      //MAKE_CARRY
    rcrb $1,%al
    adcb _PSW_B,_PSW_B  //STORE_FLAGS_SPC (SPC_FLAG_C)
    testb %al,%al
    STORE_FLAGS_SPC (SPC_FLAG_NZ)
    SET_BYTE_SPC
    jmp SPC_BACK
.endm

.macro SPC_MOV_reg_reg dest,src,score=_
ALIGNC
SPC_MOV_\dest\score\src:
    incl _NativePC
    movb __SPC_\src,%al
    testb %al,%al
    movb %al,__SPC_\dest
    STORE_FLAGS_SPC (SPC_FLAG_NZ)
    jmp SPC_BACK
.endm

.macro SPC_MOV_mem_reg addr,reg,score=_
ALIGNC
SPC_MOV_\addr\score\reg:
    ADDR_\addr
    movb __SPC_\reg,%al
    SET_BYTE_SPC
    jmp SPC_BACK
.endm

.macro SPC_MOV_mem_A addr
    SPC_MOV_mem_reg \addr,A
.endm

.macro SPC_MOV_mem_X addr
    SPC_MOV_mem_reg \addr,X
.endm

.macro SPC_MOV_mem_Y addr
    SPC_MOV_mem_reg \addr,Y
.endm

.macro SPC_MOV_reg_mem reg,addr
    ADDR_\addr
    GET_BYTE_SPC
    testb %al,%al
    movb %al,\reg
    STORE_FLAGS_SPC (SPC_FLAG_NZ)
    jmp SPC_BACK
.endm

.macro SPC_MOV_A_mem addr
ALIGNC
SPC_MOV_A_\addr:
    SPC_MOV_reg_mem R_A,\addr
.endm

.macro SPC_MOV_X_mem addr
ALIGNC
SPC_MOV_X_\addr:
    SPC_MOV_reg_mem R_X,\addr
.endm

.macro SPC_MOV_Y_mem addr
ALIGNC
SPC_MOV_Y_\addr:
    SPC_MOV_reg_mem R_Y,\addr
.endm

.macro SPC_PUSH_reg reg
ALIGNC
SPC_PUSH_\reg:
    incl _NativePC
    movb __SPC_\reg,%al
    PUSH_B
    jmp SPC_BACK
.endm

.macro SPC_POP_reg reg
ALIGNC
SPC_POP_\reg:
    incl _NativePC
    POP_B
    movb %al,__SPC_\reg
    jmp SPC_BACK
.endm
